#!/usr/bin/env python3

# ----------------------------------------------------------------------------
# Copyright (c) 2020--, Qiyun Zhu.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

"""Functions for matching reads and genes using an ordinal system.
"""

from collections import defaultdict
from itertools import chain
from operator import itemgetter

from .align import infer_align_format, assign_parser


def ordinal_mapper(fh, coords, fmt=None, n=1000000, th=0.8, prefix=False):
    """Read an alignment file and match reads and genes in an ordinal system.

    Parameters
    ----------
    fh : file handle
        Alignment file to parse.
    coords : dict
        Gene coordinates table.
    fmt : str, optional
        Alignment file format.
    n : int, optional
        Number of lines per chunk.
    th : float
        Minimum threshold of overlap length : alignment length for a match.
    prefix : bool
        Prefix gene IDs with Nucleotide IDs.

    See Also
    --------
    align.plain_mapper

    Yields
    ------
    tuple of str
        Query queue.
    dictview of set of str
        Subject(s) queue.
    """
    # determine file format
    fmt, head = (fmt, []) if fmt else infer_align_format(fh)

    # assign parser for given format
    parser = assign_parser(fmt)

    # choose match function depending on whether prefix
    match_func = match_read_gene_pfx if prefix else match_read_gene

    # cached list of query Ids for reverse look-up
    # gene Ids are unique, but read Ids can have duplicates (i.e., one read is
    # mapped to multiple loci on a genome), therefore an incremental integer
    # here replaces the original read Id as its identifer
    rids = []
    rid_append = rids.append

    # cached map of read to alignment length
    lenmap = defaultdict(dict)

    # cached map of read to coordinates
    locmap = defaultdict(list)

    def flush():
        """Match reads in current chunk with genes from all nucleotides.

        Returns
        -------
        tuple of str
            Query queue.
        dictview of set of str
            Subject(s) queue.
        """
        # master read-to-gene(s) map
        res = defaultdict(set)

        for nucl, loci in locmap.items():

            # merge and sort coordinates
            # question is to merge an unsorted list into a sorted one
            # Python's built-in timsort algorithm is efficient at this
            try:
                queue = sorted(coords[nucl] + loci)

            # it's possible that no gene was annotated on the nucleotide
            except KeyError:
                continue

            # map reads to genes using the core algorithm
            for read, gene in match_func(queue, lenmap[nucl], th, nucl):

                # merge read-gene pairs to the master map
                res[rids[read]].add(gene)

        # return matching read Ids and gene Ids
        return res.keys(), res.values()

    this = None  # current query Id
    target = n   # target line number at end of current chunk

    # parse alignment file
    for i, line in enumerate(chain(iter(head), fh)):

        # parse current alignment line
        try:
            query, subject, _, length, start, end = parser(line)[:6]
        except (TypeError, IndexError):
            continue

        # skip if length is not available
        if not length:
            continue

        # when query Id changes and chunk limits has been reached
        if query != this and i >= target:

            # flush: match currently cached reads with genes and yield
            yield flush()

            # re-initiate read Ids, length map and location map
            rids = []
            rid_append = rids.append
            lenmap, locmap = defaultdict(dict), defaultdict(list)

            # next target line number
            target = i + n

        # append read Id, alignment length and location
        idx = len(rids)
        rid_append(query)
        lenmap[subject][idx] = length
        locmap[subject].extend((
            (start, True, False, idx),
            (end,  False, False, idx)))

        this = query

    # final flush
    yield flush()


def ordinal_parser(fh, parser):
    """Alignment parsing functionalities stripped from for `ordinal_mapper`.

    Parameters
    ----------
    fh : file handle
        Alignment file to parse.
    parser : callable
        Function to parse alignment lines of certain format.

    Returns
    -------
    list of str
        Read Ids in same order as in alignment file.
    defaultdict of dict of int
        Map of read indices to alignment lengths per nucleotide.
    defaultdict of list of (int, bool, bool, str)
        Flattened list of read coordinates per nucleotide.

    See Also
    --------
    ordinal_mapper
    read_gene_coords
    .tests.test_ordinal.OrdinalTests.test_ordinal_parser

    Notes
    -----
    This is a dummy function only for test and demonstration purpose but not
    called anywhere in the program. See its unit test for details.

    The data structure is central to the algorithm. See `read_gene_coords` for
    details.
    """
    rids = []
    lenmap = defaultdict(dict)
    locmap = defaultdict(list)

    for line in fh:
        try:
            query, subject, _, length, start, end = parser(line)[:6]
        except (TypeError, IndexError):
            continue
        idx = len(rids)
        rids.append(query)
        lenmap[subject][idx] = length
        locmap[subject].extend((
            (start, True, False, idx),
            (end,  False, False, idx)))

    return rids, lenmap, locmap


def read_gene_coords(fh, sort=False):
    """Read coordinates of genes on genomes.

    Parameters
    ----------
    fh : file handle
        Gene coordinates file.
    sort : bool, optional
        Whether sort gene coordinates.

    Returns
    -------
    dict of list of (int, bool, bool, str)
        Flattened list of gene coordinates per nucleotide.
            Coordinate (nt).
            Whether start (True) or end (False).
            Whether gene (True) or read (False).
            Identifier of gene.

    See Also
    --------
    match_read_gene

    Notes
    -----
    This data structure is central to this algorithm. Starting and ending
    coordinates of each gene are separated and flattened into a sorted list.
    which enables only one round of list traversal for the entire set of genes
    plus reads.
    """
    res = {}
    queue_extend = None
    for line in fh:

        # ">" or "#" indicates genome (nucleotide) name
        c0 = line[0]
        if c0 in '>#':

            # double ">" or "#" indicates genome name, which serves as
            # a super group of subsequent nucleotide names; to be ignored
            if line[1] != c0:
                nucl = line[1:].strip()
                res[nucl] = []
                queue_extend = res[nucl].extend
        else:
            x = line.rstrip().split('\t')

            # start and end are based on genome (nucleotide), not gene itself
            try:
                start, end = sorted((int(x[1]), int(x[2])))
            except (IndexError, ValueError):
                raise ValueError(
                    f'Cannot extract coordinates from line: "{line}".')
            idx = x[0]
            queue_extend(((start, True, True, idx),
                          (end,  False, True, idx)))

    # sort gene coordinates per nucleotide
    if sort:
        sortkey = itemgetter(0)
        for nucl, queue in res.items():
            res[nucl] = sorted(queue, key=sortkey)
    return res


def whether_prefix(coords):
    """Determine whether gene IDs should be prefixed with nucleotide IDs.

    Parameters
    ----------
    coords : dict
        Gene coordinates table.

    Returns
    -------
    bool
        Whether gene IDs should be prefixed.

    See Also
    --------
    read_gene_coords

    Notes
    -----
    It is based on a simple mechanism which checks whether there are duplicate
    gene IDs, and if so, all gene IDs should be prefixed to avoid confusion.
    """
    genes = {}
    for nucl, queue in coords.items():
        for coord, is_start, is_gene, gid in queue:
            if gid not in genes:
                genes[gid] = is_start
            elif genes[gid] == is_start:
                return True
    return False


def match_read_gene(queue, lens, th, pfx=None):
    """Associate reads with genes based on a sorted queue of coordinates.

    Parameters
    ----------
    queue : list of tuple
        Sorted list of elements.
        (loc, is_start, is_gene, id)
    lens : dict
        Read-to-alignment length map.
    th : float
        Threshold for read/gene overlapping fraction.
    pfx : str, optional
        Placeholder for compatibility with match_read_gene_pfx

    Yields
    ------
    int
        Read index.
    str
        Gene ID.

    See Also
    --------
    match_read_gene_pfx
    read_gene_coords
    .tests.test_ordinal.OrdinalTests.test_match_read_gene

    Notes
    -----
    This algorithm is the core of this module. It uses a flattened, sorted
    list to store starting and ending coordinates of both genes and reads.
    Only one round of traversal (O(n)) of this list is needed to accurately
    find all gene-read matches.

    Refer to its unit test `test_match_read_gene` for an actual example and
    illustration.
    """
    genes = {}  # current genes
    reads = {}  # current reads

    # cache method references
    reads_items = reads.items
    genes_items = genes.items

    # walk through flattened queue of reads and genes
    for loc, is_start, is_gene, idx in queue:
        if is_gene:

            # when a gene starts, added to current genes
            if is_start:
                genes[idx] = loc

            # when a gene ends,
            else:

                # check current reads
                for rid, rloc in reads_items():

                    # is a match if read/gene overlap is long enough
                    if loc - max(genes[idx], rloc) + 1 >= lens[rid] * th:
                        yield rid, idx

                # remove it from current genes
                del(genes[idx])

        # the same for reads
        else:
            if is_start:
                reads[idx] = loc
            else:
                for gid, gloc in genes_items():
                    if loc - max(reads[idx], gloc) + 1 >= lens[idx] * th:
                        yield idx, gid
                del(reads[idx])


def match_read_gene_pfx(queue, lens, th, pfx):
    """Associate reads with genes based on a sorted queue of coordinates.

    Parameters
    ----------
    queue : list of tuple
        Sorted list of elements.
        (loc, is_start, is_gene, id)
    lens : dict
        Read-to-alignment length map.
    th : float
        Threshold for read/gene overlapping fraction.
    pfx : str
        Prefix to append to gene IDs.

    Yields
    ------
    int
        Read index.
    str
        Gene ID.

    See Also
    --------
    match_read_gene

    Notes
    -----
    This function is identical to `match_read_gene`, except for that it adds a
    prefix (usually a nucleotide ID) to each gene ID.
    """
    genes, reads = {}, {}
    reads_items, genes_items = reads.items, genes.items
    for loc, is_start, is_gene, idx in queue:
        if is_gene:
            if is_start:
                genes[idx] = loc
            else:
                for rid, rloc in reads_items():
                    if loc - max(genes[idx], rloc) + 1 >= lens[rid] * th:
                        yield rid, f'{pfx}_{idx}'
                del(genes[idx])
        else:
            if is_start:
                reads[idx] = loc
            else:
                for gid, gloc in genes_items():
                    if loc - max(reads[idx], gloc) + 1 >= lens[idx] * th:
                        yield idx, f'{pfx}_{gid}'
                del(reads[idx])



#!/usr/bin/env python3

# ----------------------------------------------------------------------------
# Copyright (c) 2020--, Qiyun Zhu.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

"""QIIME 2 plugin for Woltka.

Notes
-----
Minimum requirement: QIIME 2 2019.1
    This is because Woltka is dependent on Python 3.6+, which was adopted by
    QIIME 2 since 2019.1.
"""

from io import StringIO

import biom
from pandas import Series
from skbio import TreeNode

from woltka.workflow import build_mapper, classify as cwf
from woltka.file import read_map_1st, read_map_many, read_map_all
from woltka.table import prep_table, calc_coverage
from woltka.biom import table_to_biom, filter_biom, collapse_biom
from woltka.tree import read_nodes, read_lineage, read_newick, fill_root
from woltka.__init__ import __name__, __version__


def classify(alignment:             str,
             target_rank:           str,
             reference_taxonomy: Series = None,
             reference_tree:   TreeNode = None,
             reference_nodes:       str = None,
             taxon_map:             str = None,
             trim_subject:         bool = False,
             gene_coordinates:      str = None,
             overlap_threshold:     int = 80,
             unique_assignment:    bool = False,
             majority_threshold:    int = None,
             above_given_rank:     bool = False,
             subject_is_okay:      bool = False,
             report_unassigned:    bool = False) -> biom.Table:
    """Classify sequences based on their alignments to references through a
    hierarchical classification system.
    """
    # validate classification system
    num_ref = len(list(filter(None.__ne__, (
        reference_taxonomy, reference_tree, reference_nodes))))
    if num_ref > 1:
        raise ValueError('Only one reference classification system can be '
                         'specified.')
    elif num_ref == 0 and target_rank != 'none':
        raise ValueError('A reference classification system must be specified '
                         f'for classification at the rank "{target_rank}".')

    # build classification hierarchy
    tree, rankdic, namedic = {}, {}, {}

    # read taxonomy
    if reference_taxonomy is not None:
        tree, rankdic = read_lineage(StringIO(reference_taxonomy.to_csv(
            sep='\t', header=False)))

    # read phylogeny
    if reference_tree is not None:
        tree = read_newick(StringIO(str(reference_tree)))

    # read taxdump
    if reference_nodes is not None:
        with open(reference_nodes, 'r') as fh:
            tree, rankdic = read_nodes(fh)

    # read taxon mapping
    if taxon_map is not None:
        with open(taxon_map, 'r') as fh:
            tree.update(read_map_1st(fh))

    # fill root
    root = fill_root(tree)

    # build mapping module
    mapper, chunk = build_mapper(
        coords_fp=gene_coordinates, overlap=overlap_threshold)

    # classify query sequences
    profile = cwf(mapper=mapper,
                  files=[alignment],
                  demux=True,
                  trimsub=trim_subject and '_',
                  tree=tree,
                  rankdic=rankdic,
                  namedic=namedic,
                  root=root,
                  ranks=[target_rank],
                  uniq=unique_assignment,
                  major=majority_threshold,
                  above=above_given_rank,
                  subok=subject_is_okay,
                  unasgd=report_unassigned,
                  chunk=chunk,
                  zippers={})[target_rank]

    # generate feature table
    table = table_to_biom(*prep_table(
        profile, rankdic=rankdic, namedic=namedic))
    table.generated_by = f'{__name__}-{__version__}'

    return table


def psfilter(table:  biom.Table,
             min_count:     int = None,
             min_percent: float = None) -> biom.Table:
    """Filter a feature table by per-sample abundance of features.
    """
    # validate parameters
    if not any((min_count, min_percent)):
        raise ValueError('Please specify either minimum count or minimum '
                         'percentage threshold.')
    if all((min_count, min_percent)):
        raise ValueError('Only one of minimum count or minimum percentage '
                         'thresholds can be specified.')
    if min_percent and min_percent >= 100:
        raise ValueError('Minimum percentage threshold must be below 100.')

    # determine threshold
    th = min_count or min_percent / 100

    # filter feature table
    table = filter_biom(table, th)
    table.generated_by = f'{__name__}-{__version__}'

    return table


def collapse(table: biom.Table,
             mapping:      str,
             normalize:   bool = False) -> biom.Table:
    """Collapse a feature table based on many-to-many mapping.
    """
    with open(mapping, 'r') as fh:
        mapping = read_map_many(fh)
    table = collapse_biom(table, mapping, normalize)
    table.generated_by = f'{__name__}-{__version__}'
    return table


def coverage(table: biom.Table,
             mapping:      str,
             threshold:    int = None,
             count:       bool = False) -> biom.Table:
    """Calculate a feature table's coverage over feature groups.
    """
    with open(mapping, 'r') as fh:
        mapping = dict(read_map_all(fh))
    table = calc_coverage(table, mapping, threshold, count)
    table = table_to_biom(*table)
    table.generated_by = f'{__name__}-{__version__}'
    return table



#!/usr/bin/env python3

# ----------------------------------------------------------------------------
# Copyright (c) 2020--, Qiyun Zhu.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

"""Functions for handling input and output files.
"""

from os import listdir
from os.path import basename, dirname, splitext, isfile, join
from shutil import which
from subprocess import Popen, PIPE
import gzip
import bz2
import lzma


zipfmts = {'.gz':   'gzip', '.gzip':   'gzip',
           '.bz2': 'bzip2', '.bzip2': 'bzip2',
           '.xz':     'xz', '.lz':       'xz', '.lzma': 'xz'}
ziplibs = {'gzip': gzip, 'bzip2': bz2, 'xz': lzma}


def openzip(fp, mode='rt'):
    """Open a regular or compressed file by matching filename extension to
    proper library.

    Parameters
    ----------
    fp : str
        Input filepath.
    mode : str, optional
        Python file mode. Default: "rt" (read as text).

    Returns
    -------
    file handle
        Text stream ready to be read.

    Notes
    -----
    This is a simple and universal solution which uses Python's built-in
    compression modules. It supports reading and writing. However it is not as
    fast as `readzip` in reading compressed files.

    See Also
    --------
    readzip
    """
    ext = splitext(fp)[1]
    zipper = getattr(ziplibs[zipfmts[ext]], 'open') if ext in zipfmts else open
    return zipper(fp, mode)


def readzip(fp, zippers=None):
    """Open a regular or compressed file by matching filename extension to
    proper library.

    Parameters
    ----------
    fp : str
        Input filepath.
    zippers : dict of bool, optional
        Available external compression programs.

    Returns
    -------
    file handle
        Text stream ready to be read.

    Notes
    -----
    This function attempts to call external compression programs to read the
    compressed file. This is typically faster than using Python's built-in
    compression modules, because it saves the overhead in the Python wrapper,
    and utilizes additional thread(s) for decompression.

    The function checks the availability of a certain compression program at
    its first use, and stores this information in the parameter `zippers` which
    is shared across the program.

    If the external compression program is not available, or disabled (when
    `parameter` is not provided), the function will call Python's built-in
    modules instead.

    In addition to this method, a pure Python method for accelerating reading
    compression files is to use the buffered reader:

    >>> with gzip.open(fp, 'rb') as fh:
    >>>     with io.BufferedReader(fh, 1048576) as buf:
    >>>         for line in buf:
    >>>             res = line.decode()
    >>>             ...

    See Also
    --------
    openzip
    """
    # filename extension
    ext = splitext(fp)[1]

    # not a compressed file
    if ext not in zipfmts:
        return open(fp, 'r')

    fmt = zipfmts[ext]

    # external programs are disabled
    if zippers is None:
        return ziplibs[fmt].open(fp, 'rt')

    # check whether specific external program exists
    if fmt not in zippers:
        zippers[fmt] = bool(which(fmt))

    # use external program
    if zippers[fmt]:
        return Popen([fmt, '-cdfq', fp], stdout=PIPE, encoding='utf-8').stdout

    # external program does not exist
    else:
        return ziplibs[fmt].open(fp, 'rt')


def file2stem(fname, ext=None):
    """Extract stem from filename.

    Parameters
    ----------
    fp : str
        Filepath.
    ext : str, optional
        Filename extension.

    Returns
    -------
    str
        Filename stem.

    Notes
    -----
    Common compression file extensions are recognized and stripped beforehand.

    There is no commonly accepted term for "filename with extension stripped".
    The term "stem" is from C++. In Python's `splitext` function, it is called
    "root". But "root" in this program describes tree root.
    """
    if ext is not None:
        if not fname.endswith(ext):
            raise ValueError('Filepath and filename extension do not match.')
        return fname[:-len(ext)]
    else:
        stem, ext = splitext(fname)
        if ext in zipfmts:
            stem, ext = splitext(stem)
        return stem


def path2stem(fp, ext=None):
    """Get filename stem from filepath.

    Parameters
    ----------
    fp : str
        Filepath.
    ext : str, optional
        Filename extension.

    Returns
    -------
    str
        Filename stem.
    """
    return file2stem(basename(fp), ext)


def stem2rank(fp):
    """Extract rank name from stem filename.

    Parameters
    ----------
    fp : str
        Filepath.
    ext : str, optional
        Filename extension.

    Returns
    -------
    str
        Filename stem.
    """
    stem = path2stem(fp)

    # find pattern "a_to_b", "a-2-b" etc.
    for sep in ('-', '_'):
        try:
            left, middle, right = stem.split(sep)
        except ValueError:
            continue
        if middle in ('to', '2'):
            return right

    # find pattern "a2b"
    try:
        left, right = stem.split('2')
    except ValueError:
        pass
    else:
        return right

    # if neither is found, return entire string
    return stem


def read_ids(fh):
    """Read a list of IDs from a file.

    Parameters
    ----------
    fh : iterable of str
        ID list.

    Returns
    -------
    list of str
        ID list.

    Notes
    -----
    Only the first column before <tab> is considered. Lines starting with "#"
    are omitted. Empty entries are discarded.
    """
    if fh is None:
        return
    res = []
    for line in fh:
        if not line.startswith('#'):
            id_ = line.strip().partition('\t')[0]
            if id_:
                res.append(id_)
    if not res:
        raise ValueError('No ID is read.')
    if len(set(res)) < len(res):
        raise ValueError('Duplicate IDs found.')
    return res


def id2file_from_dir(dir_, ext=None, ids=None):
    """Generate an ID-to-file map from a directory.

    Parameters
    ----------
    dir_ : str
        Directory containing files.
    ext : str, optional
        Filename extension.
    ids : iterable of str, optional
        ID list.

    Returns
    -------
    dict
        ID-to-file map.

    Raises
    ------
    ValueError
        Multiple files have the same stem filename.

    Notes
    -----
    Only top-level directory is searched. Only files but not subdirectories are
    considered.
    """
    res = {}
    for fname in listdir(dir_):
        if isfile(join(dir_, fname)):
            try:
                id_ = file2stem(fname, ext)
            except ValueError:
                continue
            if ids and id_ not in ids:
                continue
            if id_ in res:
                raise ValueError(f'Ambiguous files for ID: "{id_}".')
            res[id_] = fname
    return res


def id2file_from_map(fp):
    """Read ID-to-file mapping from a file.

    Parameters
    ----------
    fp : str
        Path to an ID-to-file mapping file (format: ID <tab> filepath).

    Returns
    -------
    list of tuple, or None
        Ordered ID-to-file map, or None if not a valid mapping file.

    Raises
    ------
    ValueError
        If any of 2nd to nth filepaths does not exist.

    Notes
    -----
    This function checks every line of a file, except for empty lines and lines
    starting with "#", to see if it suffices the format of sample ID <tab>
    filepath.

    The filepath must point to an existing file. The function first checks the
    filepath as absolute, or relative to the current directory. If not found,
    it then searches the same directory as the mapping file. Therefore, instead
    of specifying full paths, one could also provide just filenames as long as
    the mapping file and the alignment files are in the same directory.

    If the filepath in the first line does not exist, the function will return
    None, assuming this is not a mapping file. However if the first filepath
    exists whereas any of the filepaths below does not exist, an error will be
    raised, reminding user of the potentially incorrect filepath.
    """
    res = []
    fdir = dirname(fp)
    with openzip(fp) as fh:
        for line in fh:
            line = line.rstrip()

            # skip empty or commented lines
            if not line or line.startswith('#'):
                continue

            # a valid map must have exactly two columns per line
            try:
                id_, file_ = line.split('\t')
            except ValueError:
                return

            # check full filepath
            if isfile(file_):
                res.append((id_, file_))
                continue

            # search same directory
            fp = join(fdir, file_)
            if isfile(fp):
                res.append((id_, fp))
                continue

            # if previous lines appear to be valid files, raise error
            if res:
                raise ValueError(f'Alignment file "{file_}" does not exist.')

            # otherwise (i.e., this is the first line), return
            return
    return res or None


def read_map_uniq(fh, sep='\t'):
    """Read a mapping file, considering unique matches only (exactly two
    columns).

    Parameters
    ----------
    fh : file handle
        Mapping file.

    Yields
    ------
    tuple of (str, str)
        Key-value pair.
    """
    for line in fh:
        key, found, value = line.partition(sep)
        if found and sep not in value:
            yield key, value.rstrip()


def read_map_1st(fh, sep='\t'):
    """Read a mapping file, considering the first two columns as key and value
    while discarding remaining columns if any.

    Parameters
    ----------
    fh : file handle
        Mapping file.

    Yields
    ------
    tuple of (str, str)
        Key-value pair.
    """
    for line in fh:
        key, found, rest = line.partition(sep)
        if found:
            value, found, rest = rest.partition(sep)
            yield key, value.rstrip()


def read_map_all(fh, sep='\t'):
    """Read a mapping file, considering all columns as key (1st) and values
    (2nd-last).

    Parameters
    ----------
    fh : file handle
        Mapping file.

    Yields
    ------
    tuple of (str, list of str)
        Key-value(s) pair.
    """
    for line in fh:
        key, found, rest = line.partition(sep)
        if found:
            yield key, rest.rstrip().split(sep)


def read_map_many(fh, sep='\t'):
    """Read many-to-many relationships from a mapping file.

    Parameters
    ----------
    fh : file handle
        Mapping file.

    Returns
    -------
    dict of list
        Key-value(s) pair.

    Notes
    -----
    This is a general solution which applies to the following formats:
    1 (one-to-one):
        source1 <tab> target1
        source2 <tab> target2
        source3 <tab> target3
        ...
    2 (one-to-many):
        source1 <tab> target1
        source2 <tab> target1 <tab> target2
        source3 <tab> target2 <tab> target3 <tab> target4
        ...
    3 (many-to-many):
        source1 <tab> target1
        source1 <tab> target2
        source2 <tab> target2
        source3 <tab> target3
        source4 <tab> target3
        ...
    """
    res = {}
    for key, values in read_map_all(fh, sep):
        res.setdefault(key, []).extend(values)
    return res


def write_readmap(fh, qryque, taxque, namedic=None):
    """Write a read map to a tab-delimited file.

    Parameters
    ----------
    fh : file handle
        Output file.
    qryque : iterable of str
        Query sequences.
    taxque : iterable of str or dict
        Taxon(a) assigned to each query.
    namedic : dict, optional
        Taxon name dictionary.
    """
    # sort subjects by count (high-to-low) then by alphabet
    def sortkey(x): return -x[1], x[0]
    for query, taxa in zip(qryque, taxque):
        row = [query]
        if isinstance(taxa, dict):
            for taxon, count in sorted(taxa.items(), key=sortkey):
                if namedic and taxon in namedic:
                    taxon = namedic[taxon]
                row.append(taxon + ':' + str(count))
        elif namedic and taxa in namedic:
            row.append(namedic[taxa])
        else:
            row.append(taxa)
        print('\t'.join(row), file=fh)



#!/usr/bin/env python3

# ----------------------------------------------------------------------------
# Copyright (c) 2020--, Qiyun Zhu.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

"""Generic utility functions that are not specific to certain bioinformatics
operations.
"""


def add_dict(dic, key, value):
    """Add a key-value pair to a dictionary while checking conflicts with
    existing ones.

    Parameters
    ----------
    dic : dict
        Dictionary to add to.
    key : hashable
        Key to be added.
    value : any
        Value to be added.

    Raises
    ------
    AssertionError
        Key already exists and its value is different.

    See Also
    --------
    update_dict

    Notes
    -----
    The behaviors of "add" and "update" (see also "update_dict") are analogous
    to Python set add() and update(), but with conflict checking.
    """
    try:
        assert dic[key] == value, f'Conflicting values found for "{key}".'
    except KeyError:
        dic[key] = value


def update_dict(dic, other):
    """Update one dictionary with another while checking conflicting items.

    Parameters
    ----------
    dic : dict
        Dictionary to be updated.
    other : dict
        Dictionary as source of new items.

    Raises
    ------
    AssertionError
        Two dictionaries have conflicting values of the same key.

    See Also
    --------
    add_dict
    sum_dict

    Notes
    -----
    If conflict-checking is not necessary, one should use Python's built-in
    method `update`.
    """
    for key, value in other.items():
        add_dict(dic, key, value)


def sum_dict(dic, other):
    """Merge one dictionary with another, adding values for common keys.

    Parameters
    ----------
    dic : dict
        Dictionary to be updated.
    other : dict
        Dictionary as source of new items.

    See Also
    --------
    update_dict
    """
    for key, value in other.items():
        dic[key] = dic.get(key, 0) + value


def intize(num):
    """Round a floating point number to an integer.

    Parameters
    ----------
    num : float
        Number to round.

    See Also
    --------
    initize_list
    initize_dict

    Notes
    -----
    Limited by Python floating point arithmetic, odds are that when ambiguous
    assignment is on, counts will be sums of fractions and they can be rounded
    either up or down when converted to integers, depending how these numbers
    are summed. For example:

    >>> data = (1/4, 1/3, 1/4, 1/3, 1/4, 1/3, 1/3, 1/3, 1/4, 1/4, 1/4, 1/3)
    >>> total = sum(data)
    >>> total, round(total)
    (3.5, 4)

    >>> total = sum(sum(data[i:i+4]) for i in range(0, len(data), 4))
    >>> total, round(total)
    (3.499999999999999, 3)

    To avoid this inconsistency, this function first checks if the count is
    close enough to a half number, and if yes, it will use the half number
    instead of the original number for rounding.

    Note: Multiple scientific computing languages such as Python and R's
    default round behavior is to round half numbers to the nearest even
    number. For example:

    >>> round(0.5)
    0
    >>> round(1.5)
    2
    """
    # first, round to the nearest half number
    near = round(num * 2) / 2

    # check if the difference is small enough:
    # yes - round the half number
    # no - round the original number
    if abs(num - near) <= 0.0000001:
        return round(near)
    else:
        return round(num)


def intize_list(lst):
    """Convert list elements to integers in place.

    Parameters
    ----------
    list : list
        Input list.

    See Also
    --------
    intize

    Notes
    -----
    Use this function instead of `initize` on large lists to save some
    function-calling overhead.
    """
    for i, element in enumerate(lst):
        near = round(element * 2) / 2
        if abs(element - near) <= 0.0000001:
            lst[i] = round(near)
        else:
            lst[i] = round(element)


def intize_dict(dic, zero=False):
    """Convert dictionary values to integers in place.

    Parameters
    ----------
    dic : dict
        Input dictionary.
    zero : bool, optional
        Whether keep zero values.

    See Also
    --------
    intize

    Notes
    -----
    Use this function instead of `initize` on large dictionaries to save some
    function-calling overhead.
    """
    todel = []
    add_todel = todel.append
    for key, value in dic.items():
        near = round(value * 2) / 2
        if abs(value - near) <= 0.0000001:
            intval = round(near)
        else:
            intval = round(value)

        # keep or skip zero values
        if intval or zero:
            dic[key] = intval
        else:
            add_todel(key)
    for key in todel:
        del dic[key]


def delnone(dic):
    """Delete None key if any from a dictionary.

    Parameters
    ----------
    dic : dict
        Input dictionary.
    """
    try:
        del dic[None]
    except KeyError:
        pass


def allkeys(dic):
    """Get all keys in a dict of dict.

    Parameters
    ----------
    dic : dict
        Input dictionary.

    Returns
    -------
    set
        Keys.
    """
    return set().union(*dic.values())


def count_list(lst):
    """Count occurrences of elements of a list and return a dictionary.

    Parameters
    ----------
    lst : list
        Input list.

    Returns
    -------
    dict
        Element-to-count map.
    """
    res = {}
    for x in lst:
        res[x] = res.get(x, 0) + 1
    return res


def last_value(lst):
    """Get last value which is not None from a list.

    Parameters
    ----------
    lst : list
        Input list.

    Returns
    -------
    scalar or None
        Last element which is not None, or None if not found.
    """
    try:
        return next(x for x in reversed(lst) if x is not None)
    except StopIteration:
        pass


def feature_count(s, sep=':'):
    """Read feature and count from a string in format of "feature:count".

    Parameters
    ----------
    s : str
        Input string.
    sep : str, optional
        Separator between feature and count (default: colon).

    Returns
    -------
    tuple of (str, int)
        Pair of feature and count.
    """
    # find first occurrence of separator from right
    feature, _, count = s.rpartition(sep)

    # if separator is found
    if feature:

        # part after separator is number
        try:
            count = int(count)

        # otherwise, entire string is feature
        except ValueError:
            return s, 1

        # count must be positive integer
        if count > 0:
            return feature, count

        # otherwise, entire string is feature
        else:
            return s, 1

    # no separator: entire string is feature
    else:
        return s, 1



#!/usr/bin/env python3

# ----------------------------------------------------------------------------
# Copyright (c) 2020--, Qiyun Zhu.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

from unittest import TestCase, main
from os import remove
from os.path import join, dirname, realpath
from shutil import rmtree
from tempfile import mkdtemp
from filecmp import cmp

import pandas as pd
from biom import load_table
from pandas.testing import assert_frame_equal

from woltka.workflow import (
    workflow, classify, parse_samples, parse_strata, build_mapper,
    prepare_ranks, build_hierarchy, assign_readmap, strip_suffix, demultiplex,
    read_strata, round_profiles, write_profiles)


class WorkflowTests(TestCase):
    def setUp(self):
        self.tmpdir = mkdtemp()
        self.datdir = join(dirname(realpath(__file__)), 'data')

    def tearDown(self):
        rmtree(self.tmpdir)

    def test_workflow(self):
        # simplest ogu workflow
        input_fp = join(self.datdir, 'align', 'bowtie2')
        output_fp = join(self.tmpdir, 'tmp.tsv')
        obs = workflow(input_fp, output_fp)['none']
        self.assertEqual(obs['S01']['G000011545'], 48)
        self.assertNotIn('G000007145', obs['S02'])
        self.assertEqual(obs['S03']['G000009345'], 640)
        self.assertTrue(cmp(output_fp, join(
            self.datdir, 'output', 'bowtie2.ogu.tsv')))
        remove(output_fp)

    def test_classify(self):
        # simplest ogu workflow
        input_fp = join(self.datdir, 'align', 'bowtie2')
        samples, files, demux = parse_samples(input_fp)
        mapper, chunk = build_mapper()
        ranks = ['none']
        obs = classify(mapper, files, samples=samples, demux=demux,
                       ranks=ranks, chunk=chunk)['none']
        self.assertEqual(obs['S01']['G000011545'], 48)
        self.assertNotIn('G000007145', obs['S02'])
        self.assertEqual(obs['S03']['G000009345'], 640)
        self.assertEqual(obs['S04']['G000240185'], 4)
        self.assertEqual(obs['S05']['G000191145'], 10)

        # complex genus/process stratification workflow
        input_fp = join(self.datdir, 'align', 'burst')
        coords_fp = join(self.datdir, 'function', 'coords.txt.xz')
        map_fps = [join(self.datdir, 'function', 'uniref', 'uniref.map.xz'),
                   join(self.datdir, 'function', 'go', 'process.tsv.xz')]
        strata_dir = join(self.datdir, 'output', 'burst.genus.map')
        samples, files, demux = parse_samples(input_fp)
        tree, rankdic, namedic, root = build_hierarchy(
            map_fps=map_fps, map_as_rank=True)
        mapper, chunk = build_mapper(coords_fp=coords_fp, overlap=80)
        stratmap = parse_strata(strata_dir, samples)
        obs = classify(
            mapper, files, samples=samples, demux=demux, tree=tree,
            rankdic=rankdic, namedic=namedic, root=root, stratmap=stratmap,
            chunk=chunk, ranks=['process'])['process']
        self.assertEqual(obs['S01'][('Thermus', 'GO:0005978')], 2)
        self.assertEqual(obs['S02'][('Bacteroides', 'GO:0006814')], 1)
        self.assertEqual(obs['S03'][('Escherichia', 'GO:0006813')], 2)
        self.assertEqual(len(obs['S04']), 39)

    def test_parse_samples(self):
        # file (assuming demultiplexed)
        fp = join(self.tmpdir, 'input.fq')
        open(fp, 'a').close()
        obs = parse_samples(fp)
        self.assertIsNone(obs[0])
        self.assertListEqual(obs[1], [fp])
        self.assertTrue(obs[2])

        # confirm demux
        obs = parse_samples(fp, demux=True)
        self.assertIsNone(obs[0])
        self.assertListEqual(obs[1], [fp])
        self.assertTrue(obs[2])

        # force non-demux
        obs = parse_samples(fp, demux=False)
        self.assertListEqual(obs[0], ['input'])
        self.assertDictEqual(obs[1], {fp: 'input'})
        self.assertFalse(obs[2])

        # provide correct extension
        obs = parse_samples(fp, ext='.fq', demux=False)
        self.assertListEqual(obs[0], ['input'])
        self.assertDictEqual(obs[1], {fp: 'input'})

        # provide wrong extension
        with self.assertRaises(ValueError) as ctx:
            parse_samples(fp, ext='.fastq', demux=False)
        self.assertEqual(str(ctx.exception), (
            'Filepath and filename extension do not match.'))

        # provide Id list (no effect since demux)
        obs = parse_samples(fp, samples='input')
        self.assertListEqual(obs[0], ['input'])
        self.assertListEqual(obs[1], [fp])

        # provide correct Id list
        obs = parse_samples(fp, samples='input', demux=False)
        self.assertListEqual(obs[0], ['input'])
        self.assertDictEqual(obs[1], {fp: 'input'})

        # provide wrong Id list
        with self.assertRaises(ValueError) as ctx:
            parse_samples(fp, samples='hello', demux=False)
        self.assertEqual(str(ctx.exception), (
            'Provided sample IDs and actual files are inconsistent.'))

        # directory (assuming per-sample)
        obs = parse_samples(self.tmpdir)
        self.assertListEqual(obs[0], ['input'])
        self.assertDictEqual(obs[1], {fp: 'input'})
        self.assertFalse(obs[2])

        # directory empty
        remove(fp)
        with self.assertRaises(ValueError) as ctx:
            parse_samples(self.tmpdir)
        self.assertEqual(str(ctx.exception), (
            'No valid file found in directory.'))

        # multiple files
        for i in range(1, 4):
            open(join(self.tmpdir, f'S{i}.sam'), 'a').close()
        obs = parse_samples(self.tmpdir)[1]
        exp = {join(self.tmpdir, f'S{i}.sam'): f'S{i}' for i in range(1, 4)}
        self.assertDictEqual(obs, exp)

        # add an irrelevant file
        fp = join(self.tmpdir, 'readme.txt')
        open(fp, 'a').close()
        obs = parse_samples(self.tmpdir)[1]
        exp[fp] = 'readme'
        self.assertDictEqual(obs, exp)

        # specify extension to target alignment files only
        del exp[fp]
        obs = parse_samples(self.tmpdir, ext='.sam')[1]
        self.assertDictEqual(obs, exp)

        # specify sample Ids
        obs = parse_samples(self.tmpdir, samples='S1,S2,S3')[1]
        self.assertDictEqual(obs, exp)

        # some samples are not found
        remove(fp)
        with self.assertRaises(ValueError) as ctx:
            parse_samples(self.tmpdir, samples='S1,S2,S4')
        self.assertEqual(str(ctx.exception), (
            'Provided sample IDs and actual files are inconsistent.'))

        # force demux
        obs = parse_samples(self.tmpdir, demux=True)
        self.assertIsNone(obs[0])
        self.assertTrue(obs[2])
        exp = [join(self.tmpdir, f'S{i}.sam') for i in range(1, 4)]
        self.assertListEqual(obs[1], exp)

        # sample Ids are ignored when demux
        obs = parse_samples(self.tmpdir, samples='S1,S2,S4', demux=True)
        self.assertListEqual(obs[0], ['S1', 'S2', 'S4'])
        exp = [join(self.tmpdir, f'S{i}.sam') for i in range(1, 4)]
        self.assertListEqual(obs[1], exp)

        # sample-to-file map
        fp = join(self.tmpdir, 'sample.list')
        with open(fp, 'w') as f:
            for i in (range(1, 4)):
                print(f'S{i}\tS{i}.sam', file=f)
        obs = parse_samples(fp)
        self.assertListEqual(obs[0], ['S1', 'S2', 'S3'])
        exp = {join(self.tmpdir, f'S{i}.sam'): f'S{i}' for i in range(1, 4)}
        self.assertDictEqual(obs[1], exp)

        # some samples only
        obs = parse_samples(fp, samples='S1,S2')
        self.assertListEqual(obs[0], ['S1', 'S2'])

        # some samples are not found
        with self.assertRaises(ValueError) as ctx:
            parse_samples(fp, samples='S1,S2,S4')
        self.assertEqual(str(ctx.exception), (
            'Provided sample IDs and actual files are inconsistent.'))
        remove(fp)

        # not a valid path
        with self.assertRaises(ValueError) as ctx:
            parse_samples('im/not/path')
        self.assertEqual(str(ctx.exception), (
            '"im/not/path" is not a valid file or directory.'))

    def test_parse_strata(self):
        # default
        for i in range(1, 4):
            open(join(self.tmpdir, f'S{i}.txt'), 'a').close()
        obs = parse_strata(self.tmpdir)
        exp = {f'S{i}': join(self.tmpdir, f'S{i}.txt') for i in range(1, 4)}
        self.assertDictEqual(obs, exp)

        # with sample Ids
        obs = parse_strata(self.tmpdir, samples=['S1', 'S2'])
        exp = {f'S{i}': join(self.tmpdir, f'S{i}.txt') for i in range(1, 3)}
        self.assertDictEqual(obs, exp)

        # sample missing
        with self.assertRaises(ValueError) as ctx:
            parse_strata(self.tmpdir, samples=['S1', 'S2', 'Sx'])
        self.assertEqual(str(ctx.exception), (
            'Cannot locate stratification files for one or more samples.'))

    def test_build_mapper(self):
        # plain
        obs = build_mapper()
        self.assertEqual(obs[0].__name__, 'plain_mapper')
        self.assertEqual(obs[1], 1000)

        # ordinal
        fp = join(self.tmpdir, 'coords.txt')
        with open(fp, 'w') as f:
            f.write('>G1\n1\t10\t20\n2\t35\t50\n')
        obs = build_mapper(fp)
        self.assertEqual(obs[0].func.__name__, 'ordinal_mapper')
        self.assertEqual(obs[1], 1000000)

        # ordinal with overlap threshold and chunk size
        obs = build_mapper(fp, overlap=75, chunk=50000)
        self.assertEqual(obs[0].func.__name__, 'ordinal_mapper')
        self.assertEqual(obs[0].keywords['th'], 0.75)
        self.assertEqual(obs[1], 50000)
        remove(fp)

    def test_prepare_ranks(self):
        # no rank
        obs = prepare_ranks()
        self.assertListEqual(obs[0], ['none'])

        # single rank
        obs = prepare_ranks('genus')
        self.assertListEqual(obs[0], ['genus'])

        # multiple ranks
        obs = prepare_ranks('phylum,genus,species')
        self.assertListEqual(obs[0], ['phylum', 'genus', 'species'])

        # missing rank
        with self.assertRaises(ValueError) as ctx:
            prepare_ranks('kingdom,phylum,class', rankdic={
                1: 'domain', 2: 'phylum', 3: 'phylum', 4: 'phylum'})
        self.assertEqual(str(ctx.exception), (
            'Ranks class, kingdom are not found in classification system.'))

        # free rank when there is tree
        obs = prepare_ranks(tree={'a': 1})
        self.assertListEqual(obs[0], ['free'])

        # with output read map directory
        obs = prepare_ranks(outmap_dir=self.tmpdir)
        self.assertTupleEqual(obs, (['none'], {'none': self.tmpdir}))

        # make multiple output read map directories
        obs = prepare_ranks('ko,mo', self.tmpdir)
        self.assertListEqual(obs[0], ['ko', 'mo'])
        self.assertDictEqual(obs[1], {'ko': join(self.tmpdir, 'ko'),
                                      'mo': join(self.tmpdir, 'mo')})
        rmtree(join(self.tmpdir, 'ko'))
        rmtree(join(self.tmpdir, 'mo'))

    def test_build_hierarchy(self):
        # names
        fp = join(self.tmpdir, 'names.dmp')
        with open(fp, 'w') as f:
            f.write('a\tHomo\nb\tPan\nc\tChimp\n')
        obs = build_hierarchy(names_fps=[fp])
        self.assertDictEqual(obs[2], {'a': 'Homo', 'b': 'Pan', 'c': 'Chimp'})
        remove(fp)

        # nodes
        fp = join(self.tmpdir, 'nodes.dmp')
        with open(fp, 'w') as f:
            f.write('a\td\nb\te\nc\td\nd\te\n')
        obs = build_hierarchy(nodes_fps=[fp])
        self.assertDictEqual(obs[0], {
            'a': 'd', 'b': 'e', 'c': 'd', 'd': 'e', 'e': 'e'})
        self.assertEqual(obs[3], 'e')

        # nodes with rank
        fp = join(self.tmpdir, 'nodes.dmp')
        with open(fp, 'w') as f:
            f.write('a\td\tlv3\nb\te\tlv3\nc\td\tlv3\nd\te\tlv2\ne\te\tlv1\n')
        obs = build_hierarchy(nodes_fps=[fp])
        self.assertDictEqual(obs[0], {
            'a': 'd', 'b': 'e', 'c': 'd', 'd': 'e', 'e': 'e'})
        self.assertDictEqual(obs[1], {'a': 'lv3', 'b': 'lv3', 'c': 'lv3',
                                      'd': 'lv2', 'e': 'lv1'})
        self.assertEqual(obs[3], 'e')
        remove(fp)

        # newick
        fp = join(self.tmpdir, 'tree.nwk')
        with open(fp, 'w') as f:
            f.write('((a,c)d,b)e;')
        obs = build_hierarchy(newick_fps=[fp])
        self.assertDictEqual(obs[0], {
            'a': 'd', 'b': 'e', 'c': 'd', 'd': 'e', 'e': 'e'})
        self.assertEqual(obs[3], 'e')
        remove(fp)

        # lineage
        fp = join(self.tmpdir, 'lineage.txt')
        with open(fp, 'w') as f:
            f.write('a\te;d\nb\te\nc\te;d\n')
        obs = build_hierarchy(lineage_fps=[fp])
        self.assertDictEqual(obs[0], {
            'a': 'e;d', 'b': 'e', 'c': 'e;d', 'e;d': 'e', 'e': 'e'})
        self.assertEqual(obs[3], 'e')

        # lineage with rank code
        with open(fp, 'w') as f:
            f.write('a\tk__e;p__d\nb\tk__e\nc\tk__e;p__d\n')
        obs = build_hierarchy(lineage_fps=[fp])
        self.assertDictEqual(obs[0], {
            'a': 'k__e;p__d', 'b': 'k__e', 'c': 'k__e;p__d',
            'k__e;p__d': 'k__e', 'k__e': 'k__e'})
        self.assertDictEqual(obs[1], {
            'k__e': 'kingdom', 'k__e;p__d': 'phylum'})
        self.assertEqual(obs[3], 'k__e')
        remove(fp)

        # rank-per-column table
        fp = join(self.tmpdir, 'columns.tsv')
        with open(fp, 'w') as f:
            f.write('#ID\tlv1\tlv2\n'
                    'a\te\td\n'
                    'b\te\t\n'
                    'c\te\td\n')
        obs = build_hierarchy(columns_fps=[fp])
        self.assertDictEqual(obs[0], {
            'a': 'd', 'b': 'e', 'c': 'd', 'd': 'e', 'e': 'e'})
        self.assertDictEqual(obs[1], {'d': 'lv2', 'e': 'lv1'})
        self.assertEqual(obs[3], 'e')
        remove(fp)

        # simple map
        fp = join(self.tmpdir, 'map.txt')
        with open(fp, 'w') as f:
            f.write('a\tBac\nb\tArc\nc\tBac\n')
        obs = build_hierarchy(map_fps=[fp])
        self.assertDictEqual(obs[0], {
            'a': 'Bac', 'b': 'Arc', 'c': 'Bac', 'Bac': '1', 'Arc': '1',
            '1': '1'})
        self.assertEqual(obs[3], '1')

        # map as rank
        obs = build_hierarchy(map_fps=[fp], map_as_rank=True)
        self.assertDictEqual(obs[1], {'Bac': 'map', 'Arc': 'map'})
        remove(fp)

    def test_strip_suffix(self):
        subs = [{'G1_1', 'G1_2', 'G2_3', 'G3'},
                {'G1_1', 'G1.3', 'G4_5', 'G4_x'}]
        obs = strip_suffix(subs, sep='_')
        exp = [{'G1', 'G2', 'G3'},
               {'G1', 'G1.3', 'G4', 'G4'}]
        self.assertListEqual(list(obs), exp)

        subs = [{'NC_123456.1_300', 'ABCD000001.20_101'}]
        obs = strip_suffix(subs, sep='_')
        exp = [{'NC_123456.1', 'ABCD000001.20'}]
        self.assertListEqual(list(obs), exp)

        subs = [{'G1.1', 'G1.2', 'G2'},
                {'G1.1', 'G1.3', 'G3_x'}]
        obs = strip_suffix(subs, sep='.')
        exp = [{'G1', 'G2'},
               {'G1', 'G3_x'}]
        self.assertListEqual(list(obs), exp)

    def test_demultiplex(self):
        # simple case
        rmap = [('S1_R1', 5),
                ('S1_R2', 12),
                ('S1_R3', 3),
                ('S2_R1', 10),
                ('S2_R2', 8),
                ('S2_R4', 7),
                ('S3_R2', 15),
                ('S3_R3', 1),
                ('S3_R4', 5)]
        obs = demultiplex(*zip(*rmap))
        exp = {'S1': [('R1',  5),
                      ('R2', 12),
                      ('R3',  3)],
               'S2': [('R1', 10),
                      ('R2',  8),
                      ('R4',  7)],
               'S3': [('R2', 15),
                      ('R3',  1),
                      ('R4',  5)]}
        self.assertEqual(obs.keys(), exp.keys())
        for s in obs:
            self.assertListEqual(list(map(tuple, obs[s])), list(zip(*exp[s])))

        # change separator, no result
        obs = demultiplex(*zip(*rmap), sep='.')
        self.assertEqual(obs.keys(), {''})
        self.assertListEqual(list(map(tuple, obs[''])), list(zip(*rmap)))

        # enforce sample Ids
        obs = demultiplex(*zip(*rmap), samples=['S1', 'S2', 'SX'])
        self.assertEqual(obs.keys(), {'S1', 'S2'})
        for s in ('S1', 'S2'):
            self.assertListEqual(list(map(tuple, obs[s])), list(zip(*exp[s])))

    def test_assign_readmap(self):
        # simple ogu assignment
        qryq = ['R1', 'R2', 'R3']
        subq = [frozenset(x) for x in [{'G1'}, {'G1', 'G2'}, {'G2', 'G3'}]]
        assigners = {}
        data = {'none': {}}
        assign_readmap(qryq, subq, data, 'none', 'S1', assigners)
        self.assertDictEqual(data['none']['S1'], {
            'G1': 1.5, 'G2': 1.0, 'G3': 0.5})

        # write read map
        data = {'none': {'S1': {}}}
        assign_readmap(qryq, subq, data, 'none', 'S1', assigners, rank2dir={
            'none': self.tmpdir})
        self.assertDictEqual(data['none']['S1'], {
            'G1': 1.5, 'G2': 1.0, 'G3': 0.5})
        fp = join(self.tmpdir, 'S1.txt')
        with open(fp, 'r') as f:
            obs = f.read().splitlines()
        exp = ['R1\tG1', 'R2\tG1:1\tG2:1', 'R3\tG2:1\tG3:1']
        self.assertListEqual(obs, exp)
        remove(fp)

        # unique and unassigned
        assigners = {}
        data = {'none': {}}
        assign_readmap(qryq, subq, data, 'none', 'S1', assigners, uniq=True,
                       unasgd=True)
        self.assertDictEqual(data['none']['S1'], {'G1': 1, 'Unassigned': 2})

        # free-rank assignment
        tree = {'G1': 'T1', 'G2': 'T1', 'G3': 'T2',
                'T1': 'T0', 'T2': 'T0', 'T0': 'T0'}
        data = {'free': {}}
        assign_readmap(qryq, subq, data, 'free', 'S1', assigners, tree=tree)
        self.assertDictEqual(data['free']['S1'], {'T0': 1, 'T1': 2})

        # fixed-rank assignment
        rankdic = {'T1': 'ko', 'T2': 'ko', 'T0': 'mo'}
        data = {'ko': {}}
        assign_readmap(qryq, subq, data, 'ko', 'S1', assigners, tree=tree,
                       rankdic=rankdic)
        self.assertDictEqual(data['ko']['S1'], {'T1': 2.5, 'T2': 0.5})

    def test_read_strata(self):
        # regular strata file
        fp = join(self.datdir, 'output', 'burst.genus.map', 'S01.txt.gz')
        obs = read_strata(fp, zippers={'gzip': False})
        self.assertEqual(len(obs), 1608)

        # file has no strata
        with self.assertRaises(ValueError) as ctx:
            read_strata(join(self.datdir, 'tree.nwk'))
        self.assertEqual(str(ctx.exception), (
            'No stratification information is found in file: tree.nwk.'))

    def test_round_profiles(self):
        # free-rank: don't round
        obs = {'free': {'S1': {'G1': 1, 'G2': 2},
                        'S2': {'G1': 3, 'G3': 2}}}
        round_profiles(obs)
        exp = {'free': {'S1': {'G1': 1, 'G2': 2},
                        'S2': {'G1': 3, 'G3': 2}}}
        self.assertEqual(obs, exp)

        # none-rank: round
        obs = {'none': {'S1': {'G1': 1.3, 'G2': 2.2},
                        'S2': {'G1': 3.0, 'G3': 2.1}}}
        round_profiles(obs)
        exp = {'none': {'S1': {'G1': 1, 'G2': 2},
                        'S2': {'G1': 3, 'G3': 2}}}
        self.assertEqual(obs, exp)

        # none-rank and unique: don't round
        obs = {'none': {'S1': {'G1': 1.5, 'G2': 0},
                        'S2': {'G1': 3.0, 'G3': 2}}}
        round_profiles(obs, uniq=True)
        exp = {'none': {'S1': {'G1': 1.5, 'G2': 0},
                        'S2': {'G1': 3.0, 'G3': 2}}}
        self.assertDictEqual(obs, exp)

        # given-rank: round
        obs = {'class': {'S1': {'G1': 1.5, 'G2': 0},
                         'S2': {'G1': 3.0, 'G3': 2}}}
        round_profiles(obs)
        exp = {'class': {'S1': {'G1': 2},
                         'S2': {'G1': 3, 'G3': 2}}}
        self.assertDictEqual(obs, exp)

        # given-rank: don't round
        obs = {'class': {'S1': {'G1': 1.5, 'G2': 0},
                         'S2': {'G1': 3.0, 'G3': 2}}}
        round_profiles(obs, uniq=False, major=0, above=True)
        exp = {'class': {'S1': {'G1': 1.5, 'G2': 0},
                         'S2': {'G1': 3.0, 'G3': 2}}}
        self.assertDictEqual(obs, exp)

    def test_write_profiles(self):
        # do nothing
        self.assertIsNone(write_profiles({}, None))

        # single profile in TSV format
        fp = join(self.tmpdir, 'output.tsv')
        data = {'none': {'S1': {'G1': 1, 'G2': 2},
                         'S2': {'G1': 3, 'G3': 2}}}
        write_profiles(data, fp)
        with open(fp, 'r') as f:
            obs = f.read().splitlines()
        exp = ['#FeatureID\tS1\tS2', 'G1\t1\t3', 'G2\t2\t0', 'G3\t0\t2']
        self.assertListEqual(obs, exp)
        remove(fp)

        # force BIOM format (even if filename ends with .tsv)
        write_profiles(data, fp, is_biom=True)
        obs = load_table(fp).to_dataframe(dense=True).astype(int)
        exp = pd.DataFrame([[1, 3], [2, 0], [0, 2]], index=['G1', 'G2', 'G3'],
                           columns=['S1', 'S2'])
        assert_frame_equal(obs, exp)

        # multiple profiles in BIOM format (default)
        data = {'none': {'S1': {'G1': 1, 'G2': 2},
                         'S2': {'G1': 3, 'G3': 2}},
                'free': {'S1': {'Ecoli': 3, 'Strep': 0},
                         'S2': {'Ecoli': 1, 'Strep': 2}}}
        write_profiles(data, self.tmpdir)
        fp = join(self.tmpdir, 'none.biom')
        obs = load_table(fp).to_dataframe(dense=True).astype(int)
        assert_frame_equal(obs, exp)
        remove(fp)
        fp = join(self.tmpdir, 'free.biom')
        obs = load_table(fp).to_dataframe(dense=True).astype(int)
        exp = pd.DataFrame([[3, 1], [0, 2]], index=['Ecoli', 'Strep'],
                           columns=['S1', 'S2'])
        assert_frame_equal(obs, exp)
        remove(fp)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

# ----------------------------------------------------------------------------
# Copyright (c) 2020--, Qiyun Zhu.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

"""Unit tests for the ordinal system, as well as a demonstration of how the
system works. See docstrings of `test_match_read_gene`.
"""

from unittest import TestCase, main
from os.path import join, dirname, realpath
from shutil import rmtree
from tempfile import mkdtemp
from io import StringIO

from woltka.file import openzip
from woltka.align import parse_b6o_line, parse_sam_line
from woltka.ordinal import (
    match_read_gene, match_read_gene_pfx, ordinal_mapper, ordinal_parser,
    read_gene_coords, whether_prefix)


class OrdinalTests(TestCase):
    def setUp(self):
        self.tmpdir = mkdtemp()
        self.datdir = join(dirname(realpath(__file__)), 'data')

    def tearDown(self):
        rmtree(self.tmpdir)

    def test_match_read_gene(self):
        """This test demonstrates how genes and reads are matched in an
        ordinal system.

        Illustration of gene (>) and read (-) locations on a genome (=):

           reads:             ---------r1---------
                                    ---------r2---------
                                        ---------r3---------
                                          ---------r4---------
                                                  ---------r5---------
                                                  ---------r6---------
          genome:  1 ====>>>>>>>>>>>g1>>>>>>>>>>>>===>>>>>>>>>>>>>>g2>> 50

           reads:             ---------r7---------
                                   ---------r8---------
                                                    ------r9------
          genome: 51 >>>>>>>>>>>===>>>>>>>>>>>>>>g3>>>>>>>>>>>>>======= 100
          --------------------------------------------------

        This function finds matching read/gene pairs given their overlap
        is no less than a fraction of alignment length (default: 80%). In this
        example, the matching pairs are:

          r1 - g1  (whole read within gene)
          r5 - g2  (17 / 20 nt within gene, exceeding threshold)
          r6 - g2  (same as above)
          r8 - g3  (whole read within gene)

        """
        # gene table
        genes = [('g1',  5, 29),
                 ('g2', 33, 61),
                 ('g3', 65, 94)]
        # read map
        reads = [('r1', 10, 29),
                 ('r2', 16, 35),
                 ('r3', 20, 39),
                 ('r4', 22, 41),
                 ('r5', 30, 49),
                 ('r6', 30, 49),  # identical
                 ('r7', 60, 79),
                 ('r8', 65, 84),
                 ('r9', 82, 95)]  # shorter

        # length map
        lens = {'r{}'.format(i): 20 for i in range(1, 10)}

        # flatten lists
        genes = [x for id_, start, end in genes for x in
                 ((start, True, True, id_),
                  (end,  False, True, id_))]
        reads = [x for id_, start, end in reads for x in
                 ((start, True, False, id_),
                  (end,  False, False, id_))]

        queue = sorted(genes + reads)

        # default (threshold = 80%)
        obs = list(match_read_gene(queue, lens, th=0.8))
        exp = [('r1', 'g1'),
               ('r5', 'g2'),
               ('r6', 'g2'),
               ('r8', 'g3')]
        self.assertListEqual(obs, exp)

        # threashold = 50%
        obs = list(match_read_gene(queue, lens, th=0.5))
        exp = [('r1', 'g1'),
               ('r2', 'g1'),
               ('r3', 'g1'),
               ('r5', 'g2'),
               ('r6', 'g2'),
               ('r7', 'g3'),
               ('r8', 'g3'),
               ('r9', 'g3')]
        self.assertListEqual(obs, exp)

    def test_match_read_gene_pfx(self):
        # same as above but adds a prefix to genes
        queue = [(1,  True,  True,  'g1'),
                 (11, True,  False, 'r1'),
                 (20, False, False, 'r1'),
                 (26, True,  False, 'r2'),
                 (35, False, False, 'r2'),
                 (50, False, True,  'g1')]
        lens = {'r1': 10, 'r2': 10}
        obs = list(match_read_gene_pfx(queue, lens, th=0.8, pfx='test'))
        exp = [('r1', 'test_g1'),
               ('r2', 'test_g1')]
        self.assertListEqual(obs, exp)

    def test_ordinal_mapper(self):
        # uses the same example as above, with some noises
        coords = read_gene_coords((
            '>n1',
            'g1	5	29',
            'g2	33	61',
            'g3	65	94',
            'gx	108	135'))
        aln = StringIO('\n'.join((
            'r1	n1	95	20	0	0	1	20	10	29	1	1',
            'r2	n1	95	20	0	0	1	20	16	35	1	1',
            'r3	n1	95	20	0	0	1	20	20	39	1	1',
            'r4	n1	95	20	0	0	20	1	22	41	1	1',
            'r5	n1	95	20	0	0	20	1	30	49	1	1',
            'rx	nx	95	20	0	0	1	20	1	20	1	1',
            'r6	n1	95	20	0	0	1	20	49	30	1	1',
            'r7	n1	95	20	0	0	25	6	79	60	1	1',
            'r8	n1	95	20	0	0	1	20	84	65	1	1',
            'r9	n1	95	20	0	0	1	20	95	82	1	1',
            'rx	nx	95	0	0	0	0	0	0	0	1	1',
            '# end of file')))
        obs = list(ordinal_mapper(aln, coords))[0]
        exp = [('r1', 'g1'),
               ('r5', 'g2'),
               ('r6', 'g2'),
               ('r8', 'g3')]
        self.assertListEqual(list(obs[0]), [x[0] for x in exp])
        self.assertListEqual(list(obs[1]), [{x[1]} for x in exp])

        # specify format
        aln.seek(0)
        obs = list(ordinal_mapper(aln, coords, fmt='b6o'))[0]
        self.assertListEqual(list(obs[0]), [x[0] for x in exp])
        self.assertListEqual(list(obs[1]), [{x[1]} for x in exp])

        # specify chunk size
        aln.seek(0)
        obs = list(ordinal_mapper(aln, coords, n=5))
        self.assertListEqual(list(obs[0][0]), [x[0] for x in exp[:2]])
        self.assertListEqual(list(obs[0][1]), [{x[1]} for x in exp[:2]])
        self.assertListEqual(list(obs[1][0]), [x[0] for x in exp[2:]])
        self.assertListEqual(list(obs[1][1]), [{x[1]} for x in exp[2:]])

        # add prefix
        aln.seek(0)
        obs = list(ordinal_mapper(aln, coords, prefix=True))[0]
        self.assertListEqual(list(obs[0]), [x[0] for x in exp])
        self.assertListEqual(list(obs[1]), [{f'n1_{x[1]}'} for x in exp])

        # specify threshold
        aln.seek(0)
        obs = list(ordinal_mapper(aln, coords, th=0.5))[0]
        exp = [('r1', 'g1'),
               ('r2', 'g1'),
               ('r3', 'g1'),
               ('r5', 'g2'),
               ('r6', 'g2'),
               ('r7', 'g3'),
               ('r8', 'g3'),
               ('r9', 'g3')]
        self.assertListEqual(list(obs[0]), [x[0] for x in exp])
        self.assertListEqual(list(obs[1]), [{x[1]} for x in exp])

    def test_ordinal_parser(self):

        # b6o (BLAST, DIAMOND, BURST, etc.)
        b6o = (
            'S1/1	NC_123456	100	100	0	0	1	100	225	324	1.2e-30	345',
            'S1/2	NC_123456	95	98	2	1	2	99	708	608	3.4e-20	270')
        parser = parse_b6o_line
        obs = ordinal_parser(b6o, parser)
        self.assertListEqual(obs[0], ['S1/1', 'S1/2'])
        self.assertDictEqual(obs[1], {'NC_123456': {0: 100, 1: 98}})
        self.assertDictEqual(obs[2], {'NC_123456': [
            (225, True, False, 0), (324, False, False, 0),
            (608, True, False, 1), (708, False, False, 1)]})

        # sam (BWA, Bowtie2, Minimap2 etc.)
        sam = (
            # SAM header to be ignored
            '@HD	VN:1.0	SO:unsorted',
            # normal, fully-aligned, forward strand
            'S1	77	NC_123456	26	0	100M	*	0	0	*	*',
            # shortened, reverse strand
            'S1	141	NC_123456	151	0	80M	*	0	0	*	*',
            # not perfectly aligned, unpaired
            'S2	0	NC_789012	186	0	50M5I20M5D20M	*	0	0	*	*',
            # unaligned
            'S2	16	*	0	0	*	*	0	0	*	*')
        parser = parse_sam_line
        obs = ordinal_parser(sam, parser)
        self.assertListEqual(obs[0], ['S1/1', 'S1/2', 'S2'])
        self.assertDictEqual(obs[1], {
            'NC_123456': {0: 100, 1: 80},
            'NC_789012': {2: 90}})
        self.assertDictEqual(obs[2], {
            'NC_123456': [(26,  True, False, 0), (125, False, False, 0),
                          (151, True, False, 1), (230, False, False, 1)],
            'NC_789012': [(186, True, False, 2), (280, False, False, 2)]})

    def test_read_gene_coords(self):
        # simple case
        tbl = ('## GCF_000123456\n',
               '# NC_123456\n',
               '1	5	384\n',
               '2	410	933\n',
               '# NC_789012\n',
               '1	912	638\n',
               '2	529	75\n')
        obs = read_gene_coords(tbl, sort=True)
        exp = {'NC_123456': [
            (5,   True, True, '1'), (384, False, True, '1'),
            (410, True, True, '2'), (933, False, True, '2')],
               'NC_789012': [
            (75,  True, True, '2'), (529, False, True, '2'),
            (638, True, True, '1'), (912, False, True, '1')]}
        self.assertDictEqual(obs, exp)

        # don't sort
        obs = read_gene_coords(tbl, sort=False)['NC_789012']
        exp = [(638, True, True, '1'), (912, False, True, '1'),
               (75,  True, True, '2'), (529, False, True, '2')]
        self.assertListEqual(obs, exp)

        # incorrect formats
        # only one column
        msg = 'Cannot extract coordinates from line:'
        with self.assertRaises(ValueError) as ctx:
            read_gene_coords(('hello',))
        self.assertEqual(str(ctx.exception), f'{msg} "hello".')
        # only two columns
        with self.assertRaises(ValueError) as ctx:
            read_gene_coords(('hello\t100',))
        self.assertEqual(str(ctx.exception), f'{msg} "hello\t100".')
        # three columns but 3rd is string
        with self.assertRaises(ValueError) as ctx:
            read_gene_coords(('hello\t100\tthere',))
        self.assertEqual(str(ctx.exception), f'{msg} "hello\t100\tthere".')

        # real coords file
        fp = join(self.datdir, 'function', 'coords.txt.xz')
        with openzip(fp) as f:
            obs = read_gene_coords(f, sort=True)
        self.assertEqual(len(obs), 107)
        obs_ = obs['G000006745']
        self.assertEqual(len(obs_), 7188)
        self.assertTupleEqual(obs_[0], (372,  True,  True, '1'))
        self.assertTupleEqual(obs_[1], (806,  False, True, '1'))
        self.assertTupleEqual(obs_[2], (816,  True,  True, '2'))
        self.assertTupleEqual(obs_[3], (2177, False, True, '2'))

    def test_whether_prefix(self):
        # gene Ids are indices
        coords = {'NC_123456': [
            (5,   True, True, '1'), (384, False, True, '1'),
            (410, True, True, '2'), (933, False, True, '2')],
                  'NC_789012': [
            (75,  True, True, '2'), (529, False, True, '2'),
            (638, True, True, '1'), (912, False, True, '1')]}
        self.assertTrue(whether_prefix(coords))

        # gene Ids are unique accessions
        coords = {'NC_123456': [
            (5,   True, True,  'NP_135792.1'),
            (384, False, True, 'NP_135792.1'),
            (410, True, True,  'NP_246801.2'),
            (933, False, True, 'NP_246801.2')],
                  'NC_789012': [
            (75,  True, True,  'NP_258147.1'),
            (529, False, True, 'NP_258147.1'),
            (638, True, True,  'NP_369258.2'),
            (912, False, True, 'NP_369258.2')]}
        self.assertFalse(whether_prefix(coords))


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

# ----------------------------------------------------------------------------
# Copyright (c) 2020--, Qiyun Zhu.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

"""Functions for classifying query sequences by assigning their subjects to a
hierarchical classification system.
"""

from operator import itemgetter
from collections import defaultdict

from .util import count_list
from .tree import find_rank, find_lca


fromkeys = dict.fromkeys


def assign_none(subs, uniq=False):
    """Assign query to subjects without using a classification system.

    Parameters
    ----------
    subs : set of str
        Subjects.
    uniq : bool, optional
        Assignment must be unique.

    Returns
    -------
    str or dict
        Unique assignment or assignment-to-count map.
    """
    try:
        sub, = subs
        return sub
    except ValueError:
        return None if uniq else fromkeys(subs, 1)


def assign_free(subs, tree, root=None, subok=False):
    """Assign query based on a classification system in a rank-free manner.

    Parameters
    ----------
    subs : set of str
        Subjects.
    tree : dict
        Hierarchical classification system.
    root : str, optional
        Root identifier.
    subok : bool, optional
        Allow assignment to subjects.

    Returns
    -------
    str or dict
        Unique assignment or assignment-to-count map.
    """
    try:
        sub, = subs
        return sub if subok else tree[sub]
    except ValueError:
        lca = find_lca(subs, tree)
        return None if lca == root else lca


def assign_rank(subs, rank, tree, rankdic, root=None, major=None, above=False,
                uniq=False):
    """Assign query to a given rank in a classification system.

    Parameters
    ----------
    subs : set of str
        Subjects.
    rank : str
        Target rank.
    tree : dict
        Hierarchical classification system.
    rankdic : dict
        Rank dictionary.
    root : str, optional
        Root identifier.
    major : float, optional
        Majority-rule assignment threshold.
    above : bool, optional
        Allow assignment above rank.
    uniq : bool, optional
        Assignment must be unique.

    Returns
    -------
    str or dict
        Unique assignment or assignment-to-count map.

    TODO
    ----
    Combine major and above.
    """
    taxa = [find_rank(x, rank, tree, rankdic) for x in subs]
    tset = set(taxa)
    if len(tset) == 1:
        return taxa[0]
    elif major:
        return majority(taxa, major)
    elif above:
        if None in tset:
            return None
        lca = find_lca(tset, tree)
        return None if lca == root else lca
    elif uniq:
        return None
    else:
        return count_list(filter(None, taxa))


def count(taxque):
    """Count occurrences of taxa in a map.

    Parameters
    ----------
    taxque : iterable of str or dict
        Taxon(a) assigned to each query.

    Returns
    -------
    defaultdict of str: int
        Taxon-to-count map.
    """
    res = defaultdict(int)
    for taxa in taxque:
        try:
            # unique match (scalar)
            res[taxa] += 1
        except TypeError:
            # multiple matches (dict of subject : count), to be normalized by
            # total match count
            k = 1 / sum(taxa.values())
            for taxon, n in taxa.items():
                res[taxon] += n * k
    return res


def count_strata(qryque, taxque, strata):
    """Stratify taxa in a map and count occurrences.

    Parameters
    ----------
    qryque : iterable of str
        Query sequences.
    taxque : iterable of str or dict
        Taxon(a) assigned to each query.
    strata : dict, optional
        Query-to-feature map for stratification.

    Returns
    -------
    defaultdict of (str, str): int
        Stratified (feature, taxon)-to-count map.
    """
    res = defaultdict(int)
    for query, taxa in zip(qryque, taxque):
        if query in strata:
            feature = strata[query]
            if isinstance(taxa, dict):
                k = 1 / sum(taxa.values())
                for taxon, n in taxa.items():
                    taxon = (feature, taxon)
                    res[taxon] += n * k
            else:
                taxon = (feature, taxa)
                res[taxon] += 1
    return res


def majority(taxa, th=0.8):
    """Select taxon from list by majority rule.

    Parameters
    ----------
    taxa : list of str
        Input taxon list.
    th : float, optional
        Threshold of majority, range = (0.5, 1.0].

    Returns
    -------
    str or None
        Selected taxon.
    """
    for taxon, n in sorted(count_list(taxa).items(), key=itemgetter(1),
                           reverse=True):
        return taxon if n >= len(taxa) * th else None



#!/usr/bin/env python3

# ----------------------------------------------------------------------------
# Copyright (c) 2020--, Qiyun Zhu.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------


"""Functions for parsing alignment / mapping files.

Notes
-----
A parser function operates on a single line in the input file.
A parser function returns a tuple of:
    str : query Id
    str : subject Id
    float : alignment score, optional
    int : alignment length, optional
    start : alignment start (5') coordinate, optional
    start : alignment end (3') coordinate, optional
"""

from collections import deque
from itertools import chain
from functools import lru_cache


def plain_mapper(fh, fmt=None, n=1000):
    """Read an alignment file in chunks and yield query-to-subject(s) maps.

    Parameters
    ----------
    fh : file handle
        Alignment file to parse.
    fmt : str, optional
        Alignment file format.
    n : int, optional
        Number of lines per chunk.

    Yields
    ------
    deque of str
        Query queue.
    deque of set of str
        Subject(s) queue.

    Notes
    -----
    The design of this function aims to couple with the extremely large size of
    typical alignment files. It reads the entire file sequentially, pauses and
    processes current cache for every _n_ lines, yields and clears cache, then
    proceeds.
    """
    # determine alignment file format
    fmt, head = (fmt, []) if fmt else infer_align_format(fh)

    # assign parser for given format
    parser = assign_parser(fmt)

    # query queue and subject(s) queue
    qryque, subque = deque(), deque()

    # pre-load method references
    qry_append, sub_append = qryque.append, subque.append

    # parse alignment file
    this = None  # current query Id
    target = n   # target line number at end of current chunk
    for i, line in enumerate(chain(iter(head), fh)):

        # parse current alignment line
        try:
            query, subject = parser(line)[:2]
        except (TypeError, IndexError):
            continue

        # add subject to subject set of the same query Id
        if query == this:
            subque[-1].append(subject)

        # when query Id changes,
        else:

            # line number has reached target
            if i >= target:

                # flush current queries and subject sets
                # this happens only when: 1) query Id changes, and 2) line
                # number has reached target, so that subjects of the same query
                # won't be separated in multiple flushes
                yield qryque, subque

                # re-initiate queues and method references
                qryque, subque = deque(), deque()
                qry_append, sub_append = qryque.append, subque.append

                # next target line number
                target = i + n

            # create new query and subject set pair
            qry_append(query)
            sub_append([subject])

            # update current query Id
            this = query

    # final flush
    yield qryque, subque


def infer_align_format(fh):
    """Guess the format of an alignment file based on content.

    Parameters
    ----------
    fh : file handle
        Input alignment file.

    Returns
    -------
    str
        Alignment file format (map, b6o or sam).
    list of str
        Lines that are read in order to infer format.

    Raises
    ------
    ValueError
        Alignment file is empty or unreadable.
    ValueError
        Alignment file format cannot be determined.

    See Also
    --------
    parse_b6o_line
    parse_sam_line

    TODO
    ----
    Currently this function guesses format only based on the first line.
    This can be more robust.
    """
    # read first line of file
    try:
        line = next(fh)

    # file is empty or not readable
    except StopIteration:
        raise ValueError('Alignment file is empty or unreadable.')

    # SAM file header
    if line.split()[0] in ('@HD', '@PG'):
        return 'sam', [line]

    # tab-delimited row
    row = line.rstrip().split('\t')

    # simple query-to-subject map
    if len(row) == 2:
        return 'map', [line]

    # BLAST standard tabular format
    if len(row) >= 12:
        if all(row[i].isdigit() for i in range(3, 10)):
            return 'b6o', [line]

    # SAM format
    if len(row) >= 11:
        if all(row[i].isdigit() for i in (1, 3, 4)):
            return 'sam', [line]

    # cannot determine
    raise ValueError('Cannot determine alignment file format.')


def assign_parser(fmt):
    """Assign parser function based on format code.

    Parameters
    ----------
    fmt : str
        Alignment file format code.

    Returns
    -------
    function
        Alignment parser function.
    """
    if fmt == 'map':  # simple map of query <tab> subject
        return parse_map_line
    if fmt == 'b6o':  # BLAST format
        return parse_b6o_line
    elif fmt == 'sam':  # SAM format
        return parse_sam_line
    else:
        raise ValueError(f'Invalid format code: "{fmt}".')


def parse_map_line(line, *args):
    """Parse a line in a simple mapping file.

    Parameters
    ----------
    line : str
        Line to parse.
    args : list, optional
        Placeholder for caller compatibility.

    Returns
    -------
    tuple of (str, str)
        Query and subject.

    Notes
    -----
    Only the first two columns are considered.
    """
    query, found, rest = line.partition('\t')
    if found:
        subject, found, rest = rest.partition('\t')
        return query, subject.rstrip()


def parse_b6o_line(line):
    """Parse a line in a BLAST tabular file (b6o).

    Parameters
    ----------
    line : str
        Line to parse.

    Returns
    -------
    tuple of (str, str, float, int, int, int)
        Query, subject, score, length, start, end.

    Notes
    -----
    BLAST tabular format:
        qseqid sseqid pident length mismatch gapopen qstart qend sstart send
        evalue bitscore

    .. _BLAST manual:
        https://www.ncbi.nlm.nih.gov/books/NBK279684/
    """
    x = line.split('\t')
    qseqid, sseqid, length, score = x[0], x[1], int(x[3]), float(x[11])
    sstart, send = sorted((int(x[8]), int(x[9])))
    return qseqid, sseqid, score, length, sstart, send


def parse_sam_line(line):
    """Parse a line in a SAM file (sam).

    Parameters
    ----------
    line : str
        Line to parse.

    Returns
    -------
    tuple of (str, str, None, int, int, int)
        Query, subject, None, length, start, end.

    Notes
    -----
    SAM format:
        QNAME, FLAG, RNAME, POS, MAPQ, CIGAR, RNEXT, PNEXT, TLEN, SEQ, QUAL,
        TAGS

    .. _Wikipedia:
        https://en.wikipedia.org/wiki/SAM_(file_format)
    .. _SAM format specification:
        https://samtools.github.io/hts-specs/SAMv1.pdf
    .. _Bowtie2 manual:
        http://bowtie-bio.sourceforge.net/bowtie2/manual.shtml#sam-output
    """
    # skip header
    if line[0] == '@':
        return

    # relevant fields
    qname, flag, rname, pos, _, cigar, _ = line.split('\t', 6)

    # skip unmapped
    if rname == '*':
        return

    # leftmost mapping position
    pos = int(pos)

    # parse CIGAR string
    length, offset = cigar_to_lens(cigar)

    # append strand to read Id if not already
    if qname[-2:] not in ('/1', '/2'):
        flag = int(flag)

        # forward strand: bit 64
        if flag & (1 << 6):
            qname += '/1'

        # reverse strand: bit 128
        elif flag & (1 << 7):
            qname += '/2'

    return qname, rname, None, length, pos, pos + offset - 1


@lru_cache(maxsize=128)
def cigar_to_lens(cigar):
    """Extract lengths from a CIGAR string.

    Parameters
    ----------
    cigar : str
        CIGAR string.

    Returns
    -------
    int
        Alignment length.
    int
        Offset in subject sequence.

    Notes
    -----
    This function significantly benefits from LRU cache because high-frequency
    CIGAR strings (e.g., "150M") are common and redundant calculations can be
    saved.
    """
    align, offset = 0, 0
    n = ''  # current step size
    for c in cigar:
        if c in 'MDIHNPSX=':
            if c in 'M=X':
                align += int(n)
            elif c in 'DN':
                offset += int(n)
            n = ''
        else:
            n += c
    return align, align + offset


def parse_kraken(line):
    """Parse a line in a Kraken mapping file.

    Parameters
    ----------
    line : str
        Line to parse.

    Returns
    -------
    tuple of (str, str)
        Query and subject.

    Notes
    -----
    Kraken2 output format:
        C/U, sequence ID, taxonomy ID, length, LCA mapping

    .. _Kraken2 manual:
        https://ccb.jhu.edu/software/kraken2/index.shtml?t=manual
    """
    x = line.split('\t')
    return (x[1], x[2]) if x[0] == 'C' else (None, None)


def parse_centrifuge(line):
    """Parse a line in a Centrifuge mapping file.

    Parameters
    ----------
    line : str
        Line to parse.

    Returns
    -------
    tuple of (str, str, int, int)
        Query, subject, score, length.

    Notes
    -----
    Centrifuge output format:
        readID, seqID, taxID, score, 2ndBestScore, hitLength, queryLength,
        numMatches

    .. _Centrifuge manual:
        https://ccb.jhu.edu/software/centrifuge/manual.shtml
    """
    if line.startswith('readID'):
        return
    x = line.split('\t')
    return x[0], x[1], int(x[3]), int(x[5])



#!/usr/bin/env python3

# ----------------------------------------------------------------------------
# Copyright (c) 2020--, Qiyun Zhu.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

from importlib import import_module

from qiime2.plugin import Plugin, Str, Bool, Int, Float, Range
from q2_types.feature_table import FeatureTable, Frequency
from q2_types.feature_data import FeatureData, Taxonomy
from q2_types.tree import Phylogeny, Rooted

from ._format import (SeqAlnMapFormat, SeqAlnMapDirFmt,
                      BLAST6OutFormat, BLAST6OutDirFmt,
                      SimpleMapFormat, SimpleMapDirFmt,
                      NCBINodesFormat, NCBINodesDirFmt,
                      GeneCoordFormat, GeneCoordDirFmt)
from ._type import SeqAlnMap, BLAST6Out, SimpleMap, NCBINodes, GeneCoordinates

from woltka import __version__
from woltka.q2.plugin import classify, psfilter, collapse, coverage


plugin = Plugin(
    name='woltka',
    version=__version__,
    website="https://github.com/qiyunzhu/woltka",
    citations=[],
    short_description='Plugin for metagenomics analysis',
    description=('This is a QIIME 2 plugin supporting various analyses of '
                 'shotgun metagenome datasets'),
    package='woltka')


plugin.register_semantic_types(
    SeqAlnMap, BLAST6Out, SimpleMap, NCBINodes, GeneCoordinates)

plugin.register_formats(SeqAlnMapFormat, SeqAlnMapDirFmt,
                        BLAST6OutFormat, BLAST6OutDirFmt,
                        SimpleMapFormat, SimpleMapDirFmt,
                        NCBINodesFormat, NCBINodesDirFmt,
                        GeneCoordFormat, GeneCoordDirFmt)

plugin.register_semantic_type_to_format(
    FeatureData[SeqAlnMap], artifact_format=SeqAlnMapDirFmt)
plugin.register_semantic_type_to_format(
    FeatureData[BLAST6Out], artifact_format=BLAST6OutDirFmt)
plugin.register_semantic_type_to_format(
    FeatureData[SimpleMap], artifact_format=SimpleMapDirFmt)
plugin.register_semantic_type_to_format(
    FeatureData[NCBINodes], artifact_format=NCBINodesDirFmt)
plugin.register_semantic_type_to_format(
    GeneCoordinates, artifact_format=GeneCoordDirFmt)

alnfmts = SeqAlnMap | BLAST6Out | SimpleMap

plugin.methods.register_function(
    function=classify,
    inputs={
        'alignment': FeatureData[alnfmts],
        'reference_taxonomy': FeatureData[Taxonomy],
        'reference_tree': Phylogeny[Rooted],
        'reference_nodes': FeatureData[NCBINodes],
        'taxon_map': FeatureData[SimpleMap],
        'gene_coordinates': GeneCoordinates
    },
    input_descriptions={
        'alignment': (
            'Multiplexed sequence alignment map to be classified. Can accept '
            'SAM, BLAST6 or simple map format.'),
        'reference_taxonomy': (
            'Reference taxonomic lineage strings.'),
        'reference_tree': (
            'Reference phylogenetic tree.'),
        'reference_nodes': (
            'Reference taxonomic nodes.'),
        'taxon_map': (
            'Mapping of subject IDs to taxon IDs.'),
        'gene_coordinates': (
            'Mapping of gene IDs to start and end positions on host genomes.')
    },
    parameters={
        'target_rank': Str,
        'overlap_threshold': Int % Range(1, 100),
        'trim_subject': Bool,
        'unique_assignment': Bool,
        'majority_threshold': Int % Range(51, 99),
        'above_given_rank': Bool,
        'subject_is_okay': Bool,
        'report_unassigned': Bool
    },
    parameter_descriptions={
        'target_rank': (
            'Classify sequences at this rank. Enter "none" to directly report '
            'subjects; enter "free" for free-rank classification.'),
        'overlap_threshold': (
            'Read/gene overlapping percentage threshold.'),
        'trim_subject': (
            'Trim subject IDs at the last underscore.'),
        'unique_assignment': (
            'One sequence can only be assigned to one classification unit, or '
            'remain unassigned if there is ambiguity. Otherwise, all candidate'
            ' units are reported and their counts are normalized.'),
        'majority_threshold': (
            'In given-rank classification, use majority rule at this '
            'percentage threshold to determine assignment when there are '
            'multiple candidates.'),
        'above_given_rank': (
            'In given-rank classification, allow assigning a sequence to '
            'a higher rank if it cannot be assigned to the current rank.'),
        'subject_is_okay': (
            'In free-rank classification, allow assigning a sequence to its '
            'direct subject, if applicable, before going up in hierarchy.'),
        'report_unassigned': (
            'Report Frequency of unassigned sequences (will be marked as '
            '"Unassigned").')
    },
    outputs=[
        ('classified_table', FeatureTable[Frequency])
    ],
    output_descriptions={
        'classified_table': (
            'The resulting table of frequencies of classification units.')
    },
    name='Flexible hierarchical sequence classifier',
    description=('Classify sequences based on their alignments to references '
                 'through a hierarchical classification system.'),
    citations=[]
)


plugin.methods.register_function(
    function=psfilter,
    inputs={'table': FeatureTable[Frequency]},
    input_descriptions={'table': 'Feature table to be filtered.'},
    parameters={'min_count': Int % Range(1, None),
                'min_percent': Float % Range(0.0, 100)},
    parameter_descriptions={
        'min_count': 'Per-sample minimum count threshold.',
        'min_percent': 'Per-sample minimum count threshold.'},
    outputs=[('filtered_table', FeatureTable[Frequency])],
    output_descriptions={'filtered_table': 'Filtered feature table.'},
    name='Per-sample feature filter',
    description=('Filter a feature table by per-sample feature abundance.'),
    citations=[]
)


plugin.methods.register_function(
    function=collapse,
    inputs={'table': FeatureTable[Frequency],
            'mapping': FeatureData[SimpleMap]},
    input_descriptions={'table': 'Feature table to be collapsed.',
                        'mapping': 'Mapping of source features to targets.'},
    parameters={'normalize': Bool},
    parameter_descriptions={'normalize': (
        'Normalize target counts by number of targets per source.')},
    outputs=[('collapsed_table', FeatureTable[Frequency])],
    output_descriptions={'collapsed_table': 'Collapsed feature table.'},
    name='Many-to-many feature mapper',
    description=('Collapse a feature table based on many-to-many feature '
                 'mapping.'),
    citations=[]
)


plugin.methods.register_function(
    function=coverage,
    inputs={'table': FeatureTable[Frequency],
            'mapping': FeatureData[SimpleMap]},
    input_descriptions={'table': 'Feature table to calculate coverage.',
                        'mapping': 'Mapping of feature groups to members.'},
    parameters={'threshold': Int % Range(1, 100),
                'count': Bool},
    parameter_descriptions={
        'threshold': ('Convert coverage to presence (1) / absence (0) data by '
                      'this percentage threshold.'),
        'count': ('Record numbers of covered features instead of percentages '
                  '(overrides threshold).')},
    outputs=[('coverage_table', FeatureTable[Frequency])],
    output_descriptions={'coverage_table': 'Feature group coverage table.'},
    name='Group coverage calculator',
    description='Calculate per-sample coverage of feature groups.',
    citations=[]
)


import_module('woltka.q2._transformer')



#!/usr/bin/env python3

# ----------------------------------------------------------------------------
# Copyright (c) 2020--, Qiyun Zhu.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

from unittest import TestCase, main
from os.path import join, dirname, realpath
from shutil import rmtree
from tempfile import mkdtemp

from woltka.tree import (
    read_names, read_nodes, read_lineage, read_newick, read_columns,
    fill_root, get_lineage, lineage_str, find_rank, find_lca)


# A small test taxon set containing 15 proteobacterial species

PROTEO = {'taxmap': {  # genome Id: taxonomy Id
    'G000008865': '562',    # Escherichia coli
    'G000195995': '28901',  # Salmonella enterica
    'G000240185': '573',    # Klebsiella pneumoniae
    'G000025565': '550',    # Enterobacter cloacae
    'G000009065': '632',    # Yersinia pestis
    'G000006745': '666',    # Vibrio cholerae
    'G000012245': '317',    # Pseudomonas syringae
    'G000191145': '48296',  # Acinetobacter pittii
    'G000195715': '520',    # Bordetella pertussis
    'G000011705': '13373',  # Burkholderia mallei
    'G000006965': '382',    # Sinorhizobium meliloti
    'G000012905': '1063',   # Rhodobacter sphaeroides
    'G000195735': '782',    # Rickettsia prowazekii
    'G000008525': '210',    # Helicobacter pylori
    'G000195755': '881'},   # Desulfovibrio vulgaris

          'phylogeny': (  # Newick format
    '((((((G000191145:1.436,G000012245:0.929)N63:0.189,((((G000195995:0.140,G0'
    '00008865:0.097)N94:0.033,(G000025565:0.150,G000240185:0.127)N95:0.027)N91'
    ':0.220,G000009065:0.266)N84:0.607,G000006745:0.737)N73:1.229)N53:0.814,(G'
    '000195715:0.707,G000011705:0.657)N46:0.835)N28:1.647,(G000195735:2.210,(G'
    '000006965:1.256,G000012905:1.058)N48:0.544)N29:1.044)N20:1.107,G000195755'
    ':2.108)N13:1.043,G000008525:3.548)N2:0.207;'),

          'taxonomy': (  # Id, parent, rank, name
    '209',     '72293',   'genus',   'Helicobacter',
    '210',     '209',     'species', 'Helicobacter pylori',
    '286',     '135621',  'genus',   'Pseudomonas',
    '317',     '286',     'species', 'Pseudomonas syringae',
    '356',     '28211',   'order',   'Rhizobiales',
    '382',     '28105',   'species', 'Sinorhizobium meliloti',
    '468',     '72274',   'family',  'Moraxellaceae',
    '469',     '468',     'genus',   'Acinetobacter',
    '506',     '80840',   'family',  'Alcaligenaceae',
    '517',     '506',     'genus',   'Bordetella',
    '520',     '517',     'species', 'Bordetella pertussis',
    '543',     '91347',   'family',  'Enterobacteriaceae',
    '547',     '543',     'genus',   'Enterobacter',
    '550',     '547',     'species', 'Enterobacter cloacae',
    '561',     '543',     'genus',   'Escherichia',
    '562',     '561',     'species', 'Escherichia coli',
    '570',     '543',     'genus',   'Klebsiella',
    '573',     '570',     'species', 'Klebsiella pneumoniae',
    '590',     '543',     'genus',   'Salmonella',
    '629',     '1903411', 'genus',   'Yersinia',
    '632',     '629',     'species', 'Yersinia pestis',
    '641',     '135623',  'family',  'Vibrionaceae',
    '662',     '641',     'genus',   'Vibrio',
    '666',     '662',     'species', 'Vibrio cholerae',
    '766',     '28211',   'order',   'Rickettsiales',
    '775',     '766',     'family',  'Rickettsiaceae',
    '780',     '775',     'genus',   'Rickettsia',
    '782',     '780',     'species', 'Rickettsia prowazekii',
    '872',     '194924',  'genus',   'Desulfovibrio',
    '881',     '872',     'species', 'Desulfovibrio vulgaris',
    '1060',    '31989',   'genus',   'Rhodobacter',
    '1063',    '1060',    'species', 'Rhodobacter sphaeroides',
    '1224',    '1224',    'phylum',  'Proteobacteria',
    '1236',    '1224',    'class',   'Gammaproteobacteria',
    '13373',   '32008',   'species', 'Burkholderia mallei',
    '28105',   '82115',   'genus',   'Sinorhizobium',
    '28211',   '1224',    'class',   'Alphaproteobacteria',
    '28216',   '1224',    'class',   'Betaproteobacteria',
    '28221',   '1224',    'class',   'Deltaproteobacteria',
    '28901',   '590',     'species', 'Salmonella enterica',
    '29547',   '1224',    'class',   'Epsilonproteobacteria',
    '31989',   '204455',  'family',  'Rhodobacteraceae',
    '32008',   '119060',  'genus',   'Burkholderia',
    '48296',   '469',     'species', 'Acinetobacter pittii',
    '72274',   '1236',    'order',   'Pseudomonadales',
    '72293',   '213849',  'family',  'Helicobacteraceae',
    '80840',   '28216',   'order',   'Burkholderiales',
    '82115',   '356',     'family',  'Rhizobiaceae',
    '91347',   '1236',    'order',   'Enterobacterales',
    '119060',  '80840',   'family',  'Burkholderiaceae',
    '135621',  '72274',   'family',  'Pseudomonadaceae',
    '135623',  '1236',    'order',   'Vibrionales',
    '194924',  '213115',  'family',  'Desulfovibrionaceae',
    '204455',  '28211',   'order',   'Rhodobacterales',
    '213115',  '28221',   'order',   'Desulfovibrionales',
    '213849',  '29547',   'order',   'Campylobacterales',
    '1903411', '91347',   'family',  'Yersiniaceae')}

# tree diagram
#
#                                 /-G000191145 Acinetobacter
#                         /N63---|
#                        |        \-G000012245 Pseudomonas
#                        |
#                        |              /-G000195995 Salmonella
#                        |           /N94
#                     /N53          |   \-G000008865 Escherichia
#                    |   |        /N91
#                    |   |       |  |   /-G000025565 Enterobacter
#                    |   |    /N84   \N95
#                    |   |   |   |      \-G000240185 Klebsiella
#                 /N28    \N73   |
#                |   |       |    \-G000009065 Yersinia
#                |   |       |
#                |   |        \-----G000006745 Vibrio
#                |   |
#             /N20   |          /-G000195715 Bordetella
#            |   |    \N46-----|
#            |   |              \-G000011705 Burkholderia
#            |   |
#            |   |          /-G000195735 Rickettsia
#         /N13    \N29-----|
#        |   |             |        /-G000006965 Sinorhizobium
#        |   |              \N48---|
# -N2----|   |                      \-G000012905 Rhodobacter
#        |   |
#        |    \---------------G000195755 Desulfovibrio
#        |
#         \---------------G000008525 Helicobacter


class TreeTests(TestCase):
    def setUp(self):
        self.tmpdir = mkdtemp()
        self.datdir = join(dirname(realpath(__file__)), 'data')
        lst_ = PROTEO['taxonomy']
        ran_ = range(0, len(lst_), 4)
        self.proteo = {
            'map': PROTEO['taxmap'],
            'tree':   dict((lst_[i], lst_[i + 1]) for i in ran_),
            'ranks':  dict((lst_[i], lst_[i + 2]) for i in ran_),
            'names':  dict((lst_[i], lst_[i + 3]) for i in ran_)}

    def tearDown(self):
        rmtree(self.tmpdir)

    def test_read_names(self):
        # simple map
        tsv = ('1	root',
               '2	Bacteria',
               '6	Azorhizobium	* notes')
        obs = read_names(tsv)
        exp = {'1': 'root', '2': 'Bacteria', '6': 'Azorhizobium'}
        self.assertDictEqual(obs, exp)

        # names.dmp
        dmp = ('2	|	Bacteria	|	Bacteria	|	scientific name	|',
               '2	|	Procaryotae	|	Procaryotae <Bacteria>	|	in-part	|',
               '2157	|	Woese et al. 1990	|		|	authority	|',
               '2157	|	Archaea	|		|	scientific name	|')
        obs = read_names(dmp)
        exp = {'2': 'Bacteria', '2157': 'Archaea'}
        self.assertDictEqual(obs, exp)

        # real names file
        fp = join(self.datdir, 'taxonomy', 'names.dmp')
        with open(fp, 'r') as f:
            obs = read_names(f)
        self.assertEqual(len(obs), 486)
        self.assertEqual(obs['356'], 'Rhizobiales')
        self.assertEqual(obs['1117'], 'Cyanobacteria')

    def test_read_nodes(self):
        # simple map
        tsv = ('1	1',
               '2	1',
               '1224	2',
               '1236	1224',
               '32066	2')
        obs0 = read_nodes(tsv)[0]
        exp0 = {'1': '1', '2': '1', '1224': '2', '1236': '1224', '32066': '2'}
        self.assertDictEqual(obs0, exp0)

        # simple map with rank
        tsv = ('1	1	no rank',
               '2	1	superkingdom',
               '1224	2	phylum',
               '1236	1224	class',
               '32066	2	phylum')
        obs0, obs1 = read_nodes(tsv)
        self.assertDictEqual(obs0, exp0)
        exp1 = {'1': 'no rank', '2': 'superkingdom', '1224': 'phylum',
                '1236': 'class', '32066': 'phylum'}
        self.assertDictEqual(obs1, exp1)

        # nodes.dmp
        dmp = ('1	|	1	|	no rank	|',
               '2	|	1	|	superkingdom	|',
               '1224	|	2	|	phylum	|',
               '1236	|	1224	|	class	|',
               '32066	|	2	|	phylum	|')
        obs0, obs1 = read_nodes(dmp)
        self.assertDictEqual(obs0, exp0)
        self.assertDictEqual(obs1, exp1)

        # real nodes file
        fp = join(self.datdir, 'taxonomy', 'nodes.dmp')
        with open(fp, 'r') as f:
            obs0, obs1 = read_nodes(f)
        self.assertEqual(len(obs0), 486)
        self.assertEqual(obs0['356'], '28211')
        self.assertEqual(obs0['1117'], '1798711')
        self.assertEqual(len(obs1), 486)
        self.assertEqual(obs1['561'], 'genus')
        self.assertEqual(obs1['562'], 'species')

    def test_read_newick(self):
        # simple tree
        nwk = '((a,b)c,(d,e)f)g;'
        exp = {'a': 'c', 'b': 'c', 'd': 'f', 'e': 'f',
               'c': 'g', 'f': 'g', 'g': 'g'}
        self.assertDictEqual(read_newick((nwk,)), exp)

        # tree with branch lengths
        nwk = '((a:0.25,b:0.12)c:2.0,(d,e:1.75)f:0.55)g;'
        self.assertDictEqual(read_newick((nwk,)), exp)

        # multi-line tree file
        nwk = (' ((a:0.25,b:0.12)c:2.0, '
               ' (d,e:1.75)f:0.55)g;    ')
        self.assertDictEqual(read_newick(nwk), exp)

        # tree without internal node Id
        nwk = '((a,b),(c,d));'
        with self.assertRaises(ValueError) as ctx:
            read_newick((nwk,))
        self.assertEqual(str(ctx.exception), 'Missing internal node ID.')

        # tree with duplicate node Ids
        nwk = '((a,b)x,(c,d)x)y;'
        with self.assertRaises(ValueError) as ctx:
            read_newick((nwk,))
        self.assertEqual(str(ctx.exception), 'Found non-unique node ID: "x".')

        # proteo tree
        obs = read_newick((PROTEO['phylogeny'],))
        exp = {'G000006965': 'N48', 'G000012905': 'N48', 'G000195735': 'N29',
               'N48': 'N29', 'G000195715': 'N46', 'G000011705': 'N46',
               'G000025565': 'N95', 'G000240185': 'N95', 'G000195995': 'N94',
               'G000008865': 'N94', 'N94': 'N91', 'N95': 'N91', 'N91': 'N84',
               'G000009065': 'N84', 'N84': 'N73', 'G000006745': 'N73',
               'G000191145': 'N63', 'G000012245': 'N63', 'N63': 'N53', 'N73':
               'N53', 'N53': 'N28', 'N46': 'N28', 'N28': 'N20', 'N29': 'N20',
               'N20': 'N13', 'G000195755': 'N13', 'N13': 'N2', 'G000008525':
               'N2', 'N2': 'N2'}
        self.assertDictEqual(obs, exp)

        # real tree file
        fp = join(self.datdir, 'tree.nwk')
        with open(fp, 'r') as f:
            obs = read_newick(f)
        self.assertEqual(len(obs), 213)
        self.assertEqual(obs['G000027165'], 'N101')
        self.assertEqual(obs['N55'], 'N46')
        self.assertEqual(obs['N2'], 'N1')
        self.assertEqual(obs['N1'], 'N1')

    def test_read_columns(self):
        # simple case
        tsv = ['#seq	k	p	c	o',
               'seq1	k1	p1	c1	o1',
               'seq2	k1	p1	c2	o2',
               'seq3	k1	p2	c3	',
               'seq4	k2		c4	o3']
        tree_obs, rankdic_obs = read_columns(iter(tsv))
        tree_exp = {'seq1': 'o1', 'o1': 'c1', 'c1': 'p1', 'p1': 'k1',
                    'seq2': 'o2', 'o2': 'c2', 'c2': 'p1',
                    'seq3': 'c3', 'c3': 'p2', 'p2': 'k1',
                    'seq4': 'o3', 'o3': 'c4', 'c4': 'k2',
                    'k1': None, 'k2': None}
        rankdic_exp = {'k1': 'k', 'k2': 'k',
                       'p1': 'p', 'p2': 'p',
                       'c1': 'c', 'c2': 'c', 'c3': 'c', 'c4': 'c',
                       'o1': 'o', 'o2': 'o', 'o3': 'o'}
        self.assertDictEqual(tree_obs, tree_exp)
        self.assertDictEqual(rankdic_obs, rankdic_exp)

        # inconsistent tree
        with self.assertRaises(ValueError) as ctx:
            read_columns(iter(tsv + ['seq5	k1	p2	c1	o1']))
        self.assertEqual(str(ctx.exception), 'Conflict at taxon "c1".')

        # inconsistent rank
        with self.assertRaises(ValueError) as ctx:
            read_columns(iter(tsv + ['seq5	k2	c4		o3']))
        self.assertEqual(str(ctx.exception), 'Conflict at taxon "c4".')

        # real rank table file
        fp = join(self.datdir, 'taxonomy', 'rank_tids.tsv')
        with open(fp, 'r') as f:
            tree, rankdic = read_columns(f)
        self.assertEqual(len(tree), 426)
        self.assertEqual(len(rankdic), 319)
        self.assertEqual(tree['562'], '561')
        self.assertEqual(tree['356'], '28211')
        self.assertEqual(rankdic['543'], 'family')
        self.assertEqual(rankdic['1224'], 'phylum')
        self.assertIsNone(tree['2'])

    def test_read_lineage(self):
        # simple case
        tsv = ('seq1	k1;p1;c1;o1',
               'seq2	k1;p2;c2;o2',
               'seq3	k1;p2;c3;',    # empty taxon in end
               'seq4	k2;p3;c1;o3',  # duplicate taxon "c1"
               'seq5	k2;;c4;o4')    # empty taxon in middle
        obs = read_lineage(tsv)[0]
        exp = {'k1':          None,
               'k2':          None,
               'k1;p1':       'k1',
               'k1;p2':       'k1',
               'k2;p3':       'k2',
               'k1;p1;c1':    'k1;p1',
               'k1;p2;c2':    'k1;p2',
               'k1;p2;c3':    'k1;p2',
               'k2;p3;c1':    'k2;p3',
               'k2;;c4':      'k2',
               'k1;p1;c1;o1': 'k1;p1;c1',
               'k1;p2;c2;o2': 'k1;p2;c2',
               'k2;p3;c1;o3': 'k2;p3;c1',
               'k2;;c4;o4':   'k2;;c4',
               'seq1':        'k1;p1;c1;o1',
               'seq2':        'k1;p2;c2;o2',
               'seq3':        'k1;p2;c3',
               'seq4':        'k2;p3;c1;o3',
               'seq5':        'k2;;c4;o4'}
        self.assertDictEqual(obs, exp)

        # simple case with rank code
        tsv = ('#OTU ID	Taxon',
               'G1	k__Bacteria; p__Firmicutes; c__Bacilli',
               'G2	k__Bacteria; p__Firmicutes; c__Clostridia',
               'G3	k__Bacteria; p__Tenericutes; c__Mollicutes',
               'G4	k__Viruses; x__Inoviridae')
        tree_obs, rankdic_obs = read_lineage(tsv)
        tree_exp = {'k__Bacteria': None,
                    'k__Viruses': None,
                    'k__Bacteria;p__Firmicutes': 'k__Bacteria',
                    'k__Bacteria;p__Tenericutes': 'k__Bacteria',
                    'k__Bacteria;p__Firmicutes;c__Bacilli':
                    'k__Bacteria;p__Firmicutes',
                    'k__Bacteria;p__Firmicutes;c__Clostridia':
                    'k__Bacteria;p__Firmicutes',
                    'k__Bacteria;p__Tenericutes;c__Mollicutes':
                    'k__Bacteria;p__Tenericutes',
                    'k__Viruses;x__Inoviridae': 'k__Viruses',
                    'G1': 'k__Bacteria;p__Firmicutes;c__Bacilli',
                    'G2': 'k__Bacteria;p__Firmicutes;c__Clostridia',
                    'G3': 'k__Bacteria;p__Tenericutes;c__Mollicutes',
                    'G4': 'k__Viruses;x__Inoviridae'}
        self.assertDictEqual(tree_obs, tree_exp)
        rankdic_exp = {'k__Bacteria': 'kingdom',
                       'k__Viruses': 'kingdom',
                       'k__Bacteria;p__Firmicutes': 'phylum',
                       'k__Bacteria;p__Tenericutes': 'phylum',
                       'k__Bacteria;p__Firmicutes;c__Bacilli': 'class',
                       'k__Bacteria;p__Firmicutes;c__Clostridia': 'class',
                       'k__Bacteria;p__Tenericutes;c__Mollicutes': 'class'}
        self.assertDictEqual(rankdic_obs, rankdic_exp)

        # real lineage file
        fp = join(self.datdir, 'taxonomy', 'lineage.txt')
        with open(fp, 'r') as f:
            tree, rankdic = read_lineage(f)
        self.assertEqual(len(tree), 426)
        self.assertEqual(len(rankdic), 319)
        self.assertEqual(tree['k__Bacteria;p__Firmicutes'], 'k__Bacteria')
        self.assertEqual(rankdic['k__Bacteria;p__Firmicutes'], 'phylum')
        self.assertIsNone(tree['k__Bacteria'])

    def test_get_lineage(self):
        # simple tree
        tsv = ('1	1',
               '2	1',
               '1224	2',
               '1236	1224',
               '28211	1224'
               '32066	2')
        tree = read_nodes(tsv)[0]
        obs = get_lineage('1236', tree)
        exp = ['1', '2', '1224', '1236']
        self.assertListEqual(obs, exp)
        self.assertIsNone(get_lineage('1234', tree))

        # proteo tree
        obs = get_lineage('561', self.proteo['tree'])  # Escherichia
        exp = ['1224', '1236', '91347', '543', '561']
        self.assertListEqual(obs, exp)
        obs = get_lineage('286', self.proteo['tree'])  # Pseudomonas
        exp = ['1224', '1236', '72274', '135621', '286']
        self.assertListEqual(obs, exp)

        # real tree file
        fp = join(self.datdir, 'taxonomy', 'nodes.dmp')
        with open(fp, 'r') as f:
            tree = read_nodes(f)[0]
        obs = get_lineage('561', tree)  # Escherichia
        exp = ['1', '131567', '2', '1224', '1236', '91347', '543', '561']
        self.assertListEqual(obs, exp)
        obs = get_lineage('286', tree)  # Pseudomonas
        exp = ['1', '131567', '2', '1224', '1236', '72274', '135621', '286']
        self.assertListEqual(obs, exp)

    def test_lineage_str(self):
        tree = self.proteo['tree']
        obs = lineage_str('561', tree)
        self.assertEqual(obs, '1236;91347;543')
        obs = lineage_str('561', tree, include_root=True)
        self.assertEqual(obs, '1224;1236;91347;543')
        obs = lineage_str('561', tree, include_self=True)
        self.assertEqual(obs, '1236;91347;543;561')
        obs = lineage_str('561', tree, self.proteo['names'])
        exp = 'Gammaproteobacteria;Enterobacterales;Enterobacteriaceae'
        self.assertEqual(obs, exp)
        obs = lineage_str('1224', tree)
        self.assertEqual(obs, '')
        obs = lineage_str('1236', tree)
        self.assertEqual(obs, '')
        obs = lineage_str('0000', tree)
        self.assertEqual(obs, '')

    def test_fill_root(self):
        # already has root
        tree = {'root': 'root',
                'left': 'root',
                'right': 'root'}
        self.assertEqual(fill_root(tree), 'root')

        # multiple crown nodes with parent as self
        tree = {'2': '2',
                '3': '3',
                '4': '2',
                '5': '4',
                '6': '3'}
        self.assertEqual(fill_root(tree), '1')
        self.assertIn('1', tree)
        self.assertEqual(tree['1'], '1')

        # multiple crown nodes with parent as None
        tree = {'1': None,
                '2': '1',
                '3': '1',
                '4': None,
                '5': '4'}
        self.assertEqual(fill_root(tree), '6')
        self.assertIn('6', tree)
        self.assertEqual(tree['1'], '6')
        self.assertEqual(tree['4'], '6')

        # multiple crown nodes with one non-existent parent
        tree = {'Proteobacteria': 'Bacteria',
                'Actinobacteria': 'Bacteria',
                'Firmicutes':     'Bacteria'}
        self.assertEqual(fill_root(tree), 'Bacteria')
        self.assertIn('Bacteria', tree)
        self.assertEqual(tree['Bacteria'], 'Bacteria')

        # multiple crown nodes with multiple non-existent parents
        tree = {'Proteobacteria': 'Bacteria',
                'Actinobacteria': 'Bacteria',
                'Firmicutes':     'Bacteria',
                'Euryarchaeota':  'Archaea',
                'Crenarchaeota':  'Archaea'}
        self.assertEqual(fill_root(tree), '1')
        self.assertIn('1', tree)
        self.assertEqual(tree['Bacteria'], '1')
        self.assertEqual(tree['Archaea'], '1')

    def test_find_rank(self):
        # simple case
        tree = {'1':     '1',
                '2':     '1',
                '1224':  '2',
                '1236':  '1224',
                '28211': '1224',
                '32066': '2'}
        ranks = {'1': 'no rank',
                 '2': 'superkingdom',
                 '1224': 'phylum',
                 '1236': 'class',
                 '28211': 'class',
                 '32066': 'phylum'}
        self.assertEqual(find_rank('1236', 'class', tree, ranks), '1236')
        self.assertEqual(find_rank('1236', 'phylum', tree, ranks), '1224')
        self.assertEqual(find_rank('32066', 'superkingdom', tree, ranks), '2')
        self.assertIsNone(find_rank('12345', 'species', tree, ranks))

        # proteo tree
        args = self.proteo['tree'], self.proteo['ranks']
        # Escherichia coli (species) => Enterobacteriaceae (family)
        self.assertEqual(find_rank('561', 'family', *args), '543')
        # Bordetella (genus) => Betaproteobacteria (class)
        self.assertEqual(find_rank('517', 'class', *args), '28216')
        # Yersinia pestis (species) => self (species)
        self.assertEqual(find_rank('632', 'species', *args), '632')
        # Rickettsiales (order) is already above genus
        self.assertIsNone(find_rank('766', 'genus', *args))
        # no rank is named as "tribe"
        self.assertIsNone(find_rank('317', 'tribe', *args))

    def test_find_lca(self):
        # simple tree
        tsv = ('1	1',
               '2	1',
               '1224	2',
               '1236	1224',
               '28211	1224',
               '32066	2',
               '2157	1')
        tree = read_nodes(tsv)[0]
        self.assertEqual(find_lca(['1236'], tree), '1236')
        self.assertEqual(find_lca(['1236', '32066'], tree), '2')
        self.assertEqual(find_lca(['1236', '2157'], tree), '1')
        self.assertEqual(find_lca(['1'], tree), '1')
        self.assertIsNone(find_lca(['1234'], tree))
        self.assertIsNone(find_lca(['1234', '1'], tree))
        self.assertIsNone(find_lca(['1', '2', '3'], tree))
        self.assertEqual(find_lca(['1', '1'], tree), '1')

        # broken tree, only for unit test coverage
        tree = {'1': '1', '2': '2'}
        self.assertEqual(find_lca(['1', '2'], tree), '1')

        # proteo tree
        # Escherichia, Enterobacter => Enterobacteriaceae
        obs, exp = ['562', '550'], '543'
        self.assertEqual(find_lca(obs, self.proteo['tree']), exp)
        # Helicobacter, Desulfovibrio => Proteobacteria
        obs, exp = ['209', '881'], '1224'
        self.assertEqual(find_lca(obs, self.proteo['tree']), exp)
        # Rickettsia, Sinorhizobium, Rhodobacter => Alphaproteobacteria
        obs, exp = ['782', '382', '1063'], '28211'
        self.assertEqual(find_lca(obs, self.proteo['tree']), exp)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

# ----------------------------------------------------------------------------
# Copyright (c) 2020--, Qiyun Zhu.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

from setuptools import setup, find_packages

import woltka.__init__ as init


classes = """
    Development Status :: 3 - Alpha
    License :: OSI Approved :: BSD License
    Topic :: Scientific/Engineering
    Topic :: Scientific/Engineering :: Bio-Informatics
    Programming Language :: Python :: 3
    Programming Language :: Python :: 3 :: Only
    Operating System :: Unix
    Operating System :: POSIX
    Operating System :: MacOS
"""

params = {
    'name':             init.__name__,
    'version':          init.__version__,
    'license':          init.__license__,
    'long_description': open('README.md').read(),
    'long_description_content_type': 'text/markdown',
    'author':           init.__author__,
    'author_email':     init.__email__,
    'url':              init.__url__,
    'install_requires': ['biom-format'],
    'classifiers':      classes.strip().split('\n    '),
    'python_requires':  '>=3.6',
    'entry_points': {
        'console_scripts': [f'{init.__name__}=woltka.cli:cli'],
        'qiime2.plugins': [f'q2-{init.__name__}=woltka.q2.plugin_setup:plugin']
    }
}

setup(**params, packages=find_packages(), include_package_data=True)



#!/usr/bin/env python3

# ----------------------------------------------------------------------------
# Copyright (c) 2020--, Qiyun Zhu.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

from unittest import TestCase, main
from os.path import join, dirname, realpath
from shutil import rmtree
from tempfile import mkdtemp

from woltka.classify import (
    assign_none, assign_free, assign_rank, count, count_strata, majority)


class ClassifyTests(TestCase):
    def setUp(self):
        self.tmpdir = mkdtemp()
        self.datdir = join(dirname(realpath(__file__)), 'data')

    def tearDown(self):
        rmtree(self.tmpdir)

    def test_assign_none(self):
        # assign to self
        self.assertEqual(assign_none({'G1'}), 'G1')

        # count subjects
        obs = assign_none({'G1', 'G2'})
        exp = {'G1': 1, 'G2': 1}
        self.assertDictEqual(obs, exp)

        # cannot assign
        obs = assign_none({'G1', 'G2', 'G3'}, uniq=True)
        self.assertIsNone(obs)

    def test_assign_free(self):
        tree = {'G1': 'T1', 'G2': 'T1', 'G3': 'T2',
                'T1': 'T0', 'T2': 'T0', 'T0': 'T0'}
        kwargs = {'tree': tree, 'root': 'T0'}

        # free-rank assignment
        obs = assign_free({'G1', 'G2'}, **kwargs)
        self.assertEqual(obs, 'T1')

        # root -> None
        obs = assign_free({'G1', 'G2', 'G3'}, **kwargs)
        self.assertIsNone(obs)

        # assign one sub to a taxon
        obs = assign_free({'G1'}, **kwargs)
        self.assertEqual(obs, 'T1')

        # assign one sub to itself
        obs = assign_free({'G1'}, **kwargs, subok=True)
        self.assertEqual(obs, 'G1')

    def test_assign_rank(self):
        tree = {'G1': 'T1', 'G2': 'T1', 'G3': 'T2',
                'T1': 'T0', 'T2': 'T0', 'T0': 'T0'}
        rankdic = {'T1': 'general', 'T2': 'general', 'T0': 'marshal'}
        kwargs = {'tree': tree, 'rankdic': rankdic, 'root': 'T0'}

        # fixed-rank assignment
        obs = assign_rank({'G1', 'G2', 'G3'}, 'marshal', **kwargs)
        self.assertEqual(obs, 'T0')
        obs = assign_rank({'G1', 'G2'}, 'general', **kwargs)
        self.assertEqual(obs, 'T1')
        obs = assign_rank({'G1'}, 'general', **kwargs)
        self.assertEqual(obs, 'T1')

        # rank that does not exist
        obs = assign_rank({'G1', 'G2', 'G3'}, 'admiral', **kwargs, uniq=True)
        self.assertIsNone(obs)

        # cannot find consensus at rank
        obs = assign_rank({'G1', 'G2', 'G3'}, 'general', **kwargs, uniq=True)
        self.assertIsNone(obs)

        # majority rule
        obs = assign_rank({'G1', 'G2', 'G3'}, 'general', **kwargs, major=0.6)
        self.assertEqual(obs, 'T1')

        # consider ambiguity
        obs = assign_rank({'G1', 'G2', 'G3'}, 'general', **kwargs)
        exp = {'T1': 2, 'T2': 1}
        self.assertDictEqual(obs, exp)

        # assign above rank
        obs = assign_rank({'G1', 'G2', 'G3'}, 'general', tree=tree,
                          rankdic=rankdic, root=None, above=True)
        self.assertEqual(obs, 'T0')

        # assign above rank but there is a None
        obs = assign_rank({'G1', 'G2', 'G3', None}, 'general', tree=tree,
                          rankdic=rankdic, root=None, above=True)
        self.assertEqual(obs, None)

    def test_count(self):
        # unique match
        taxque = ['a', 'a', 'b', 'c', None, 'b', None]
        obs = count(taxque)
        exp = {'a': 2, 'b': 2, 'c': 1, None: 2}
        self.assertDictEqual(obs, exp)

        # multiple matches
        taxque = [{'a': 1, 'b': 2, 'c': 3},
                  {'a': 2, 'b': 5},
                  {'d': 4}]
        obs = count(taxque)
        exp = {'a': 1 / 6 + 2 / 7,
               'b': 2 / 6 + 5 / 7,
               'c': 3 / 6,
               'd': 4 / 4}
        for key in obs:
            self.assertAlmostEqual(obs[key], exp[key])

    def test_count_strata(self):
        # unique match
        strata = {'seq1': 'Ecoli',
                  'seq2': 'Ecoli',
                  'seq3': 'Cdiff',
                  'seq4': 'Strep',
                  'seq5': 'Strep',
                  'seq6': 'Ecoli'}
        matches = {'seq1': 'ligase',
                   'seq2': 'polymerase',
                   'seq3': 'nuclease',
                   'seq4': 'ligase',
                   'seq5': 'nuclease',
                   'seq6': 'ligase',
                   'seq0': 'nothing'}
        obs = count_strata(matches.keys(), matches.values(), strata)
        exp = {('Ecoli',     'ligase'): 2,
               ('Ecoli', 'polymerase'): 1,
               ('Cdiff',   'nuclease'): 1,
               ('Strep',     'ligase'): 1,
               ('Strep',   'nuclease'): 1}
        self.assertDictEqual(obs, exp)

        # multiple matches
        strata['seq7'] = 'Ecoli'
        matches['seq7'] = {'polymerase': 3, 'ligase': 1}
        obs = count_strata(matches.keys(), matches.values(), strata)
        exp = {('Ecoli',     'ligase'): 2.25,
               ('Ecoli', 'polymerase'): 1.75,
               ('Cdiff',   'nuclease'): 1,
               ('Strep',     'ligase'): 1,
               ('Strep',   'nuclease'): 1}
        self.assertDictEqual(obs, exp)

    def test_majority(self):
        obs = majority([1, 1, 1, 2, 2], th=0.6)
        self.assertEqual(obs, 1)
        obs = majority([1, 1, 1, 2, 2], th=0.8)
        self.assertIsNone(obs)
        obs = majority([1, 2, 3])
        self.assertIsNone(obs)
        obs = majority([])
        self.assertIsNone(obs)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

# ----------------------------------------------------------------------------
# Copyright (c) 2020--, Qiyun Zhu.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

"""Functions for manipulating tables (profiles).
"""

from functools import reduce
from collections import defaultdict
from operator import add
from biom import Table, load_table

from .util import allkeys, update_dict, intize_list
from .tree import lineage_str
from .file import openzip
from .biom import (
    table_to_biom, biom_to_table, write_biom, filter_biom, round_biom,
    biom_add_metacol, collapse_biom)


def prep_table(profile, samples=None, tree=None, rankdic=None, namedic=None,
               name_as_id=False):
    """Convert a profile into data, features and samples, as well as metadata
    if applicable, which can be further converted into a TSV file, a BIOM table
    or a Pandas DataFrame.

    Parameters
    ----------
    profile : dict
        Profile data.
    samples : list, optional
        Ordered sample ID list.
    tree : dict, optional
        Taxonomic tree, to inform "Lineage" column.
    rankdic : dict, optional
        Rank dictionary, to inform "Rank" column.
    namedic : dict, optional
        Taxon name dictionary, to inform "Name" column.
    name_as_id : bool, optional
        Replace feature IDs with names. It applies to row headers and "Lineage"
        column, and removes "Name" column.

    Returns
    -------
    list of list
        Data (2D array of values).
    list
        Features (rows, Pandas index, or BIOM observation IDs).
    list
        Samples (columns, Pandas columns, or BIOM sample IDs).
    list of dict
        Metadata (extra columns, or BIOM observation metadata).

    Examples
    --------
    Convert output to a BIOM table:
    >>> import biom
    >>> args = profile, samples, tree, rankdic, namedic, name_as_id
    >>> table = biom.Table(prep_table(*args))

    Convert output to a Pandas DataFrame (data only):
    >>> import pandas as pd
    >>> data, features, samples, metadata = prep_table(*args)
    >>> df = pd.DataFrame(data, features, samples)

    Convert output to a Pandas DataFrame (data and metadata):
    >>> df = pd.concat([pd.DataFrame(data, features, samples),
                        pd.DataFrame.from_records(metadata, features)], axis=1)

    See Also
    --------
    write_tsv
    .biom.table_to_biom
    .workflow.write_profiles

    Notes
    -----
    Optionally, three metadata columns, "Name", "Rank" and "Lineage" will be
    appended to the table.

    A feature will be dropped if all its values are zero. However, samples will
    not be dropped even when empty.

    A stratified feature will be printed as "stratum|feature".
    """
    # determine range and order of samples
    samples = [x for x in samples if x in profile] if samples else sorted(
        profile)

    # determine metadata columns
    namecol = namedic and not name_as_id or None
    metacols = tuple(filter(None, (
        namecol and 'Name', rankdic and 'Rank', tree and 'Lineage')))

    notnone = None.__ne__
    features, data, metadata = [], [], []

    # sort features in alphabetical order
    for key in sorted(allkeys(profile)):

        # determine cell values (feature counts)
        datum = [profile[x][key] if key in profile[x] else 0 for x in samples]
        if not any(datum):
            continue
        data.append(datum)

        # determine feature Id
        stratum, taxon = key if isinstance(key, tuple) else (None, key)
        name = namedic[taxon] if namedic and taxon in namedic else None
        feature = name if name_as_id and name else taxon
        feature = f'{stratum}|{feature}' if stratum else feature
        features.append(feature)

        # determine metadata
        metadatum = dict(zip(metacols, filter(notnone, (
            namecol and (name or ''),
            rankdic and (rankdic[taxon] if taxon in rankdic else ''),
            tree and lineage_str(
                taxon, tree, namedic if name_as_id else None)))))
        metadata.append(metadatum)

    return data, features, samples, metadata


def read_table(fp):
    """Read a feature table (profile) from an input file, while automatically
    recognizing its format (BIOM or TSV).

    Parameters
    ----------
    fp : str
        Input filepath.

    Returns
    -------
    biom.Table or tuple
        Table that was read.
    str
        Format of table ('biom' or 'tsv)

    Raises
    ------
    ValueError
        Cannot read table.
    """
    with open(fp, 'r') as fh:
        try:
            table = read_tsv(fh)
        except UnicodeDecodeError:
            try:
                table = load_table(fp)
            except (TypeError, UnicodeDecodeError):
                raise ValueError(
                    'Input file cannot be parsed as BIOM or TSV format.')
            else:
                return table, 'biom'
        else:
            return table, 'tsv'


def write_table(table, fp, is_biom=None):
    """Write a feature table (profile) to an output file, while automatically
    matching format (BIOM or TSV).

    Parameters
    ----------
    table : biom.Table or tuple
        Table to be written.
    fp : str
        Output filepath.
    is_biom : bool, optional
        Output BIOM or TSV format. If not specified, will automatically
        determine based on output filename.
    """
    if is_biom is not False and fp.endswith('.biom'):
        if isinstance(table, tuple):
            table = table_to_biom(*table)
        write_biom(table, fp)
    else:
        if isinstance(table, Table):
            table = biom_to_table(table)
        with openzip(fp, 'wt') as fh:
            write_tsv(table, fh)


def read_tsv(fh):
    """Read a tab-delimited file into table components.

    Parameters
    ----------
    fh : file handle
        Output file.

    Returns
    -------
    list of list
        Profile data.
    list
        Feature IDs (row names).
    list
        Sample IDs (column names).
    list of dict
        Feature metadata (extra columns).

    Raises
    ------
    ValueError
        Input table file is empty.
    ValueError
        Input table file has no sample.
    """
    # get header (1st line)
    try:
        header = next(fh).rstrip('\r\n').split('\t')
    except StopIteration:
        raise ValueError('Input table file is empty.')

    # extract metadata columns
    header, metacols = strip_metacols(header)

    # extract sample Ids
    samples = header[1:]
    if not samples:
        raise ValueError('Input table file has no sample.')

    # parse table body
    data, features, metadata = [], [], []
    width = len(header)
    for line in fh:
        row = line.rstrip('\r\n').split('\t')
        feature = row[0]
        features.append(feature)
        data.append(list(map(int, row[1:width])))
        metadata.append(dict(zip(metacols, row[width:])))

    return data, features, samples, metadata


def write_tsv(table, fh):
    """Write table components to a tab-delimited file.

    Parameters
    ----------
    table : tuple
        Table components to write.
    fh : file handle
        Output file.

    Notes
    -----
    The output table will have columns as samples and rows as features.
    Optionally, three metadata columns, "Name", "Rank" and "Lineage" will be
    appended to the right of the table.
    """
    data, features, samples, metadata = table

    # metadata columns
    metacols = next(iter(metadata), {}).keys() if metadata else None

    # table header
    header = ['#FeatureID', '\t'.join(samples)]
    if metacols:
        header.append('\t'.join(metacols))
    print(*header, sep='\t', file=fh)

    # table body
    for i, feature in enumerate(features):
        row = [feature, '\t'.join(map(str, data[i]))]
        if metacols:
            row.append('\t'.join(metadata[i].values()))
        print(*row, sep='\t', file=fh)


def strip_metacols(header, cols=['Name', 'Rank', 'Lineage']):
    """Extract ordered metadata columns from the right end of a table header.

    Parameters
    ----------
    header : list
        Table header.
    cols : list, optional
        Candidate columns in order.

    Returns
    -------
    list
        Table header with metadata columns stripped.
    list
        Extracted metadata columns.

    Notes
    -----
    Metadata columns can be whole or a subset of the provided candidates, but
    they must appear in the given order at the right end of the table header.
    Unordered and duplicated columns as well as columns mixed within samples
    will not be extracted.
    """
    res = []
    for col in reversed(header):
        try:
            cols = cols[slice(cols.index(col))]
        except ValueError:
            break
        else:
            res.append(col)
    return header[:-len(res) or None], res[::-1]


def table_shape(table):
    """Get feature and sample counts of a table.

    Parameters
    ----------
    table : biom.Table, or tuple of (list, list, list, list)
        Table to count (data, features, samples, metadata).

    Returns
    -------
    int
        Number of features.
    int
        Number of samples.
    """
    if isinstance(table, Table):
        return table.shape
    return len(table[1]), len(table[2])


def filter_table(table, th):
    """Filter a table by per-sample count or percentage threshold.

    Parameters
    ----------
    table : biom.Table, or tuple of (list, list, list, list)
        Table to filter (data, features, samples, metadata).
    th : float
        Per-sample minimum abundance threshold. If >= 1, this value is an
        absolute count; if < 1, it is a fraction of sum of counts.

    Returns
    -------
    biom.Table, or tuple of (list, list, list, list)
        Filtered table (data, features, samples, metadata).
    """
    # redirect to BIOM module
    if isinstance(table, Table):
        return filter_biom(table, th)

    # filtering function to apply to each column
    def f(datum, th):
        bound = th if th >= 1 else sum(datum) * th
        return [0 if x < bound else x for x in datum]

    # transpose data array, apply function to each column (now row), then
    # transpose back
    res = ([], [], table[2], [])
    for i, datum in enumerate(zip(*(f(x, th) for x in zip(*table[0])))):
        if not any(datum):
            continue
        res[0].append(list(datum))
        res[1].append(table[1][i])
        res[3].append(table[3][i])

    return res


def merge_tables(tables):
    """Merge multiple tables into one.

    Parameters
    ----------
    tables : list
        Table to merge.

    Returns
    -------
    biom.Table, or tuple of (list, list, list, list)
        Merged table.
    """
    # merge BIOM tables using built-in method
    if all(isinstance(x, Table) for x in tables):
        return reduce(lambda a, b: a.merge(b), tables)

    data = defaultdict(lambda: defaultdict(int))
    metadata = {}
    for table in tables:
        if isinstance(table, Table):
            table = biom_to_table(table)

        # update metadata
        try:
            update_dict(metadata, dict(zip(table[1], table[3])))
        except AssertionError:
            raise ValueError('Conflicting metadata found in tables.')

        # update data
        for feature, datum in zip(table[1], table[0]):
            for sample, value in zip(table[2], datum):
                data[sample][feature] += value

    res = prep_table(data)
    return res[0], res[1], res[2], [metadata[x] for x in res[1]]


def round_table(table):
    """Round table data to integers and remove all-zero rows in place.

    Parameters
    ----------
    table : biom.Table, or tuple of (list, list, list, list)
        Table to round (data, features, samples, metadata).

    See Also
    --------
    .util.intize
        Rationale for this rounding method.
    """
    # redirect to BIOM module
    if isinstance(table, Table):
        round_biom(table)
        return

    # round table data
    todel = []
    for i, datum in enumerate(table[0]):
        intize_list(datum)
        if not any(datum):
            todel.append(i)

    # remove empty rows
    for i in reversed(todel):
        del(table[0][i])
        del(table[1][i])
        del(table[3][i])


def add_metacol(table, dic, name, missing=''):
    """Add a metadata column to a table in place based on a dictionary.

    Parameters
    ----------
    table : biom.Table, or tuple of (list, list, list, list)
        Table to add metadata column (data, features, samples, metadata).
    dict : dict
        Metadata column (feature-to-value mapping).
    name : str
        Metadata column name.
    missing : any type, optional
        Default value if not found in dictionary.
    """
    # redirect to BIOM module
    if isinstance(table, Table):
        biom_add_metacol(table, dic, name, missing=missing)
        return

    # add metadata column
    for feature, metadatum in zip(*(table[1], table[3])):
        metadatum[name] = dic.get(feature, missing)


def collapse_table(table, mapping, normalize=False):
    """Collapse a table by many-to-many mapping.

    Parameters
    ----------
    table : biom.Table, or tuple of (list, list, list, list)
        Table to collapse (data, features, samples, metadata).
    mapping : dict of list of str
        Source-to-target(s) mapping.
    normalize : bool, optional
        Whether normalize per-target counts by number of targets per source.

    Returns
    -------
    biom.Table, or tuple of (list, list, list, list)
        Collapsed table.

    Notes
    -----
    Metadata will not be retained in the collapsed table.
    """
    # redirect to BIOM module
    if isinstance(table, Table):
        return collapse_biom(table, mapping, normalize)

    # collapse table
    samples = table[2]
    width = len(samples)
    res = defaultdict(lambda: [0] * width)
    for datum, feature in zip(*table[:2]):
        if feature not in mapping:
            continue
        targets = mapping[feature]
        if normalize:
            k = 1 / len(targets)
            datum = [x * k for x in datum]
        for target in targets:
            res[target] = list(map(add, res[target], datum))

    # reformat table
    res = list(res.values()), list(res.keys()), samples, [dict() for _ in res]

    # round table
    if normalize:
        round_table(res)
    return res


def calc_coverage(table, mapping, th=None, count=False):
    """Calculate coverages of table over feature groups based on group
    memberships.

    Parameters
    ----------
    table : biom.Table, or tuple of (list, list, list, list)
        Input table (data, features, samples, metadata).
    mapping : dict of list of str
        Group-to-members mapping.
    th : int or float, optional
        Convert coverages to presence / absence data based on this threshold
        (range: (0, 100]).
    count : bool, optional
        Record number of covered features instead of percentage (overrides
        `th`).

    Returns
    -------
    tuple of (list, list, list, list)
        Coverage table.

    Notes
    -----
    Empty groups will cause ZeroDivision error, but upstream workflow already
    removes this possibility.
    """
    if isinstance(table, Table):
        table = biom_to_table(table)
    # data, features, samples, metadata
    data = [[] for x in table[2]]
    for row, feature in zip(table[0], table[1]):
        for i, value in enumerate(row):
            # check presence
            if value > 0:
                data[i].append(feature)
    data = list(map(set, data))
    covers, groups = [], []
    for group, members in mapping.items():
        members = set(members)
        total = len(members)
        row = []
        for present in data:
            # count covered features
            covered = len(present.intersection(members))
            if count:
                row.append(covered)
                continue
            # calculate coverage
            cover = 100 * covered / total
            # determine presence / absence
            if th:
                row.append(int(cover >= th))
                continue
            # retain percentage
            row.append(round(cover, 3))
        # keep group when at least one sample has coverage
        if any(row):
            covers.append(row)
            groups.append(group)
    return covers, groups, table[2], [{} for x in range(len(groups))]
